# pylint: disable=C0103, C0114, R0913
from __future__ import annotations
from pathlib import Path
import json
import urllib.request
import warnings

from smufolib import config
from smufolib.utils import error, normalizers
from smufolib.utils._annotations import JsonDict

CONFIG = config.load()


class Request:
    """Send HTTP or filesystem request.

    A fallback path (e.g., a filesystem path to the same file), may be specified in case
    of connection failure when the primary path is a URL.

    An optional warning may be raised in the event of a fallback.

    :param path: Primary URL or filepath.
    :param fallback: Fallback filepath to use if `path` raises
        :class:`urllib.error.URLError`.
    :param encoding: File text encoding. See :func:`open` for details. Defaults to
        :confval:`request.encoding`.
    :param warn: Warn if URLError is raised before fallback request. Defaults to
        :confval:`request.warn`.

    """

    @classmethod
    def classes(cls, decode: bool = True) -> JsonDict | str | None:
        """Retrieve `classes` metadata from configured paths.

        This method attempts to load metadata from the path specified in
        :confval:`metadata.paths.classes`, falling back to
        :confval:`metadata.fallbacks.classes` if necessary.

        :param decode: If :obj:`True`, return parsed JSON data; otherwise, return raw
            response. Defaults to :obj:`True`.

        """
        return cls._getMetadata("classes", decode=decode)

    @classmethod
    def glyphnames(cls, decode: bool = True) -> JsonDict | str | None:
        """Retrieve `glyphnames` metadata from configured paths.

        This method attempts to load metadata from the path specified in
        :confval:`metadata.paths.glyphnames`, falling back to
        :confval:`metadata.fallbacks.glyphnames` if necessary.

        :param decode: If :obj:`True`, return parsed JSON data; otherwise, return raw
            response. Defaults to :obj:`True`.

        """
        return cls._getMetadata("glyphnames", decode=decode)

    @classmethod
    def ranges(cls, decode: bool = True) -> JsonDict | str | None:
        """Retrieve `ranges` metadata from configured paths.

        This method attempts to load metadata from the path specified in
        :confval:`metadata.paths.ranges`, falling back to
        :confval:`metadata.fallbacks.ranges` if necessary.

        :param decode: If :obj:`True`, return parsed JSON data; otherwise, return raw
            response. Defaults to :obj:`True`.

        """
        return cls._getMetadata("ranges", decode=decode)

    @classmethod
    def font(cls, decode: bool = True) -> JsonDict | str | None:
        """Retrieve `font` metadata from configured paths.

        This method attempts to load metadata from the path specified in the
        :confval:`metadata.paths.font`, falling back to
        :confval:`metadata.fallbacks.font` if necessary.

        :param decode: If :obj:`True`, return parsed JSON data; otherwise, return raw
            response. Defaults to :obj:`True`.

        """
        return cls._getMetadata("font", decode=decode)

    @classmethod
    def _getMetadata(cls, filename: str, decode: bool = True) -> JsonDict | str | None:
        path = CONFIG["metadata.paths"][filename]
        fallback = CONFIG["metadata.fallbacks"][filename]
        request = cls(path, fallback)
        return request.json() if decode else request.text

    def __init__(
        self,
        path: Path | str | None = None,
        fallback: Path | str | None = None,
        encoding: str = CONFIG["request"]["encoding"],
        warn: bool = CONFIG["request"]["warn"],
    ) -> None:
        self._path = path
        self._fallback = fallback
        self._encoding = encoding
        self._warn = warn

    def __repr__(self):
        return (
            f"<{self.__class__.__name__} '{self.path}' "
            f"('{self.fallback}') at {id(self)}>"
        )

    def json(self) -> JsonDict | None:
        """Parse request as JSON.

        :raises json.JSONDecodeError: If the response cannot be parsed as JSON.
        :raises TypeError: If the raw data is None.

        """
        if self.text is None:
            return None
        return json.loads(self.text)

    def _readPreferredSource(self) -> bytes | None:
        if self.path is None and self.fallback is None:
            return None

        try:
            return self._readFromURL()
        # TypeError guards against path being None
        except (urllib.error.URLError, TypeError):
            if self.fallback:
                return self._readFromFallback()
            raise
        # ValueError confirms that path is an assumed filepath (non-URL)
        except ValueError:
            return self._readFromPath()

    def _readFromURL(self) -> bytes:
        # Read data from URL.
        if self.path is None:
            raise TypeError(
                error.generateTypeError(self.path, (str, Path, Request), "Request.path")
            )
        try:
            with urllib.request.urlopen(self.path) as raw:
                return raw.read()
        except urllib.error.URLError as exc:
            return self._handleURLError(exc)

    def _readFromFallback(self) -> bytes:
        # Read data from fallback file.
        if self.fallback is None:
            raise TypeError(
                error.generateTypeError(
                    self.fallback, (str, Path, Request), "Request.fallback"
                )
            )

        with open(self.fallback, "rb") as raw:
            return raw.read()

    def _readFromPath(self) -> bytes:
        # Read data from path.
        if self.path is None:
            raise TypeError(
                error.generateTypeError(self.path, (str, Path, Request), "Request.path")
            )
        with open(self.path, "rb") as raw:
            return raw.read()

    def _handleURLError(self, exc: urllib.error.URLError) -> bytes:
        # Handle URL error during online request.
        if not self.fallback:
            raise urllib.error.URLError(
                error.generateErrorMessage("urlError", url=self.path)
            ) from exc

        if self._warn:
            warnings.warn(
                error.generateErrorMessage("urlError", url=self.path), error.URLWarning
            )

        with open(self.fallback, "rb") as fallback_file:
            return fallback_file.read()

    @property
    def path(self) -> str | None:
        """Primary URL or filepath as string.

        :raises TypeError: If the path cannot be normalized.

        """
        return normalizers.normalizeRequestPath(self._path, "path")

    @path.setter
    def path(self, value: Path | str | None) -> None:
        self._path = normalizers.normalizeRequestPath(value, "path")

    @property
    def fallback(self) -> str | None:
        """Fallback URL or filepath as string.

        :raises TypeError: If the fallback path cannot be normalized.

        """
        return normalizers.normalizeRequestPath(self._fallback, "fallback")

    @fallback.setter
    def fallback(self, value: str | None) -> None:
        self._fallback = normalizers.normalizeRequestPath(value, "fallback")

    @property
    def encoding(self) -> str:
        """File text encoding."""
        return self._encoding

    @encoding.setter
    def encoding(self, value: str):
        self._encoding = value

    @property
    def content(self) -> bytes | None:
        """The raw response data as bytes.

        This property is read-only.

        """
        return self._readPreferredSource()

    @property
    def text(self) -> str | None:
        """The raw response data as text.

        This property is read-only.

        """
        raw = self._readPreferredSource()
        if raw is None:
            return None
        return raw.decode(self.encoding)


def writeJson(
    filepath: Path | str,
    source: JsonDict,
    encoding: str = CONFIG["request"]["encoding"],
) -> None:
    """Writes JSON data to filepath.

    :param filepath: Path to target file.
    :param source: JSON data source.
    :param encoding: File text encoding. See :func:`open` for details. Defaults to
        :confval:`request.encoding`.
    :raises TypeError: If `filepath` is not the expected type.
    :raises ValueError: If there is an error serializing the JSON data or if `filepath`
        does not have a ``.json`` exctension.
    :raises FileNotFoundError: If the specified `filepath` cannot be found.
    :raises OSError: If there are any issues opening or writing to the file.
    :raises UnsupportedOperation: If the operation is not supported.

    """
    error.validateType(filepath, (Path, str), "filepath")
    if not str(filepath).endswith(".json"):
        raise ValueError(
            error.generateErrorMessage(
                "missingExtension", objectName="filepath", extension=".json"
            )
        )
    try:
        with open(filepath, "w", encoding=encoding) as outfile:
            json.dump(source, outfile, indent=4, sort_keys=False)
    except (TypeError, ValueError) as e:
        raise ValueError("serializationError") from e



# pylint: disable=C0114
import defcon
from fontParts.base.font import BaseFont
from fontParts.fontshell.font import RFont
from smufolib.objects.layer import Layer
from smufolib.objects.smufl import Smufl


class Font(RFont):
    """SMufoLib environment implementation of :class:`fontParts.base.BaseFont`.

    :param pathOrObject: The source used to initialize the font.

        - If :obj:`None`, a new, empty font will be created.
        - If a :class:`str`, it is interpreted as a path to an existing UFO file.
        - If an unwrapped native font object (e.g., a :class:`defcon.Font`) or a
          :class:`fontParts.base.BaseFont`, it will be wrapped.

        .. versionadded:: 0.6.0

            Support for initializing from an existing :class:`fontParts.base.BaseFont`
            instance.

    Examples:

        >>> from smufolib import Font
        >>> font = Font("path/to/MyFont.ufo")  # doctest: +SKIP

        >>> from fontParts.fontshell import RFont
        >>> font = Font(RFont())  # doctest: +SKIP

    """

    # pylint: disable=too-few-public-methods

    layerClass = Layer

    def __init__(self, pathOrObject: str | defcon.Font | BaseFont | None = None):
        if isinstance(pathOrObject, BaseFont):
            try:
                pathOrObject = pathOrObject.naked()
            except AttributeError as exc:
                raise TypeError(
                    "Expected 'pathOrObject' to be of type str, "
                    "defcon.Font or BaseFont with naked()."
                ) from exc

        super().__init__(pathOrObject)

    @property
    def smufl(self) -> Smufl:
        """Font-specific instance of :class:`~smufolib.objects.smufl.Smufl`."""
        return Smufl(font=self)



#!/usr/bin/env python3
# pylint: disable=C0301
"""Generate font metadata JSON file.

This script generates a full featured metadata `JSON
<https://www.json.org/json-en.html>`_ file for SMuFL fonts from metadata sources and
user specifications, including the following metadata sections:

    - :smufl:`mandatory (fontName and fontVersion) <specification/font-specific-metadata.html>`
    - :smufl:`size <specification/font-specific-metadata.html>`
    - :smufl:`engravingDefaults <specification/engravingdefaults.html>`
    - :smufl:`glyphAdvanceWidths <specification/glyphadvancewidths.html>`
    - :smufl:`glyphsWithAnchors <specification/glyphswithanchors.html>`
    - :smufl:`glyphsWithAlternates <specification/glyphswithalternates.html>`
    - :smufl:`glyphBBoxes <specification/glyphbboxes.html>`
    - :smufl:`ligatures <specification/ligatures.html>`
    - :smufl:`sets <specification/sets.html>`
    - :smufl:`optionalGlyphs <specification/optionalglyphs.html>`

Sections are automatically added depending on the font's scope and assigned attribute
values.

.. tip::

    As a starting point, identification and engraving defaults
    attributes may be set automatically with the :mod:`~bin.importID`
    and :mod:`~bin.calculateEngravingDefaults` scripts respectively.

Glyphs outside the SMuFL range as well as unencoded or unnamed glyphs, are automatically
skipped.

The script requires SMufoLib to be installed in its execution environment. It can be
used from the command line or as a Python module. See the :ref:`generate-metadata-cli`
and :ref:`generate-metadata-python` sections below for usage details.

"""

from __future__ import annotations
from typing import Any
import argparse
import collections
from pathlib import Path

from tqdm import tqdm

from smufolib import Font, config
from smufolib.request import Request, writeJson
from smufolib import cli
from smufolib.utils import stdUtils
from smufolib.utils.scriptUtils import normalizeFont as _normalizeFont
from smufolib.utils.scriptUtils import normalizeJsonDict as _normalizeJsonDict
from smufolib.utils.scriptUtils import normalizeRequest as _normalizeRequest
from smufolib.utils.scriptUtils import normalizeTargetPath as _normalizeTargetPath

# Type aliases
JsonDict = dict[str, Any]
SetsTemplate = dict[str, dict[str, str]]

# Configuration
CONFIG = config.load()

# Parameter defaults
FONT_DATA = Request(
    CONFIG["metadata.paths"]["font"],
    CONFIG["metadata.fallbacks"]["font"],
)
VERBOSE = False

# pylint: disable=C0103


def generateMetadata(
    font: Font | Path | str,
    targetPath: str | Path,
    fontData: str | Path | Request = FONT_DATA,
    verbose: bool = VERBOSE,
) -> None:
    """Generate font metadata JSON file (Python API).

    :param font: Object or path to targetPath :class:`~smufolib.objects.font.Font`.
    :param targetPath: Target directory for Metadata JSON file.
    :param fontData: Request for or path to reference font metadata file.
        Defaults to :class:`smufolib.request.Request` passing
        :confval:`metadata.paths.font` and :confval:`metadata.fallbacks.font`.
    :param verbose: Make output verbose. Defaults to :obj:`False`.
    :raises TypeError: If any parameter value is not the expected type.
    :raises FileNotFoundError: If `targetPath` does not exist.

    """
    print("Starting...")

    targetPath = _normalizeTargetPath(targetPath)
    font = _normalizeFont(font)
    fontDataJson = _normalizeJsonDict(_normalizeRequest(fontData).json())
    font.smufl.spaces = True

    metadata = _compileMetadata(font, fontDataJson, verbose)

    filename = Path(targetPath).resolve() / f"{font.smufl.name}.json"
    stdUtils.verbosePrint(f"\nWriting metadata to: '{filename}'", verbose)
    writeJson(filename, metadata)

    print("\nDone.")


def main() -> None:
    """Command line entry point."""
    args = _parseArgs()
    generateMetadata(
        args.font, args.targetPath, fontData=args.fontData, verbose=args.verbose
    )


def _compileMetadata(font: Font, fontData: JsonDict, verbose: bool) -> JsonDict:
    # Build SMuFL-specific metadata dictionary.
    # pylint: disable=R0912, R0914
    stdUtils.verbosePrint("\nCompiling font metadata...", verbose)

    metadata: collections.defaultdict[str, JsonDict] = collections.defaultdict(dict)

    keys = ("fontName", "fontVersion", "designSize", "sizeRange", "engravingDefaults")

    for key in keys:
        attribute = key[4:].lower() if "font" in key else key
        value = getattr(font.smufl, attribute)
        if attribute == "engravingDefaults":
            value = value.items()
        elif value is None:
            continue
        metadata[key] = value

    sets, setGlyphs = {}, []
    setsTemplate = _getSetsTemplate(fontData)

    stdUtils.verbosePrint("\nCompiling metadata for glyphs:", verbose)
    for glyph in font if verbose else tqdm(font):
        if glyph is None or not stdUtils.validateAttr(
            glyph, ("name", "unicode", "smufl.isMember")
        ):
            stdUtils.verbosePrint("\tSkipping invalid glyph:", verbose, glyph)
            continue
        stdUtils.verbosePrint(f"\t{glyph}", verbose)

        s = glyph.smufl
        metadata["glyphAdvanceWidths"][s.name] = s.advanceWidth
        metadata["glyphBBoxes"][s.name] = s.bBox

        if s.alternates:
            metadata["glyphsWithAlternates"][s.name] = {"alternates": s.alternates}

        if s.anchors:
            metadata["glyphsWithAnchors"][s.name] = s.anchors

        if s.isLigature:
            metadata["ligatures"][s.name] = {
                "codepoint": s.codepoint,
                "componentGlyphs": s.componentNames,
                "description": s.description,
            }

        if s.isOptional:
            metadata["optionalGlyphs"][s.name] = {
                "codepoint": s.codepoint,
                "description": s.description,
                "classes": s.classes,
            }

        if s.isSet:
            sets[s.suffix] = setsTemplate[s.suffix]
            glyphInfo = {
                "alternateFor": s.base.smufl.name,
                "codepoint": s.codepoint,
                "description": s.description,
                "name": s.name,
            }
            setGlyphs.append(glyphInfo)
            metadata["sets"] = sets
            metadata["sets"][s.suffix]["glyphs"] = setGlyphs

    return metadata


def _getSetsTemplate(fontData: JsonDict) -> SetsTemplate:
    # Extract sets template from bravura_metadata.json.
    sets: dict[str, dict[str, str]] = {}

    for key, values in fontData["sets"].items():
        sets[key] = {}
        for subKey, value in values.items():
            if subKey not in ("description", "type"):
                continue
            sets[key][subKey] = value
    return sets


def _parseArgs() -> argparse.Namespace:
    # Parse command line arguments and options.
    parser = cli.commonParser(
        "font",
        "targetPath",
        description=stdUtils.getSummary(generateMetadata.__doc__),
        fontData=FONT_DATA,
        verbose=VERBOSE,
    )
    return parser.parse_args()


if __name__ == "__main__":
    main()



# pylint: disable=C0114, C0103, W0212, W0221
from __future__ import annotations
from typing import TYPE_CHECKING, cast
from collections.abc import Callable

from fontParts.base.base import BaseObject
from smufolib import config
from smufolib.utils import error, normalizers
from smufolib.utils.rulers import DISPATCHER, ENGRAVING_DEFAULTS_MAPPING
from smufolib.utils._annotations import (
    CollectionType,
    EngravingDefaultsInput,
    EngravingDefaultsReturn,
)

if TYPE_CHECKING:  # pragma: no cover
    from smufolib.objects.smufl import Smufl
    from smufolib.objects.font import Font
    from smufolib.objects.glyph import Glyph
    from smufolib.objects.layer import Layer

# Type aliases
EngravingDefaultsDictInput = dict[str, EngravingDefaultsInput | None]
EngravingDefaultsDictReturn = dict[str, EngravingDefaultsReturn | None]

AUTO = config.load()["engravingDefaults"]["auto"]

#: Names  and descriptions of engraving defaults as specified in the SMuFL standard.
ENGRAVING_DEFAULTS_ATTRIBUTES: dict[str, str] = {
    "textFontFamily": "Preferred text font families for pairing with this music font.",
    "staffLineThickness": "Thickness of each staff line.",
    "stemThickness": "Thickness of a stem.",
    "beamThickness": "Thickness of a beam.",
    "beamSpacing": "Distance between primary and secondary beams.",
    "legerLineThickness": "Thickness of a leger line.",
    "legerLineExtension": "Extension length of a leger line beyond the notehead.",
    "slurEndpointThickness": "Thickness at the end of a slur.",
    "slurMidpointThickness": "Thickness at the midpoint of a slur.",
    "tieEndpointThickness": "Thickness at the end of a tie.",
    "tieMidpointThickness": "Thickness at the midpoint of a tie.",
    "thinBarlineThickness": "Thickness of a thin barline.",
    "thickBarlineThickness": "Thickness of a thick barline.",
    "dashedBarlineThickness": "Thickness of a dashed barline.",
    "dashedBarlineDashLength": "Length of dashes in a dashed barline.",
    "dashedBarlineGapLength": "Gap length between dashes in a dashed barline.",
    "barlineSeparation": "Distance between multiple barlines when locked together.",
    "thinThickBarlineSeparation": "Distance between thin and thick barlines when locked together.",
    "repeatBarlineDotSeparation": "Horizontal distance between dots and barline in repeats.",
    "bracketThickness": "Thickness of bracket lines grouping staves.",
    "subBracketThickness": "Thickness of sub-bracket lines for related staves.",
    "hairpinThickness": "Thickness of crescendo/diminuendo hairpins.",
    "octaveLineThickness": "Thickness of dashed lines used for octave indications.",
    "pedalLineThickness": "Thickness of lines used for piano pedaling.",
    "repeatEndingLineThickness": "Thickness of brackets indicating repeat endings.",
    "arrowShaftThickness": "Thickness of arrow shafts.",
    "lyricLineThickness": "Thickness of lyric extension lines for melismas.",
    "textEnclosureThickness": "Thickness of boxes drawn around text instructions.",
    "tupletBracketThickness": "Thickness of brackets around tuplet numbers.",
    "hBarThickness": "Thickness of the H-bar in a multi-bar rest.",
}


class EngravingDefaults(BaseObject):
    """SMuFL engraving default settings.

    This object contains properties and methods pertained to SMuFL's
    :smufl:`engravingDefaults <specification/engravingdefaults.html>` metadata
    structure, defining recommended defaults for line widths etc., according to the
    specification.

    The :class:`EngravingDefaults` object is essentially a :class:`dict` with each
    engraving default setting exposed as a read/write property.

    .. versionchanged:: 0.5.0

        If a value is unassigned (or explicitly set to :obj:`None`), the attribute will
        be calculated automatically from the corresponding glyph in the font, provided
        that glyph exists and :attr:`auto` is enabled. See
        :data:`.ENGRAVING_DEFAULTS_MAPPING` for a complete list of attributes and their
        default corresponding glyphs and assigned ruler functions.

    .. tip::

        Optionally, values may be explicitly set using glyph-based calculations provided
        by the :mod:`~bin.calculateEngravingDefaults` script.

    :param smufl: Parent :class:`~smufolib.objects.smufl.Smufl` object.
    :param auto: Whether to calculate engraving defaults automatically. Defaults to
        :confval:`engravingDefaults.auto` configuration.

    This object is typically accessed through a font's Smufl metadata interface:

        >>> engravingDefaults = font.smufl.engravingDefaults
        >>> glyph = font["uniE050"]
        >>> engravingDefaults = glyph.smufl.engravingDefaults

    It may also be instantiated independently and assigned to a font later:

        >>> engravingDefaults = EngravingDefaults()  # doctest: +SKIP

    """

    # Stubs to improve linting support.
    arrowShaftThickness: int | float
    barlineSeparation: int | float
    beamSpacing: int | float
    beamThickness: int | float
    bracketThickness: int | float
    dashedBarlineDashLength: int | float
    dashedBarlineGapLength: int | float
    dashedBarlineThickness: int | float
    hairpinThickness: int | float
    hBarThickness: int | float
    legerLineExtension: int | float
    legerLineThickness: int | float
    lyricLineThickness: int | float
    octaveLineThickness: int | float
    pedalLineThickness: int | float
    repeatBarlineDotSeparation: int | float
    repeatEndingLineThickness: int | float
    slurEndpointThickness: int | float
    slurMidpointThickness: int | float
    staffLineThickness: int | float
    stemThickness: int | float
    subBracketThickness: int | float
    textEnclosureThickness: int | float
    thickBarlineThickness: int | float
    thinBarlineThickness: int | float
    thinThickBarlineSeparation: int | float
    tieEndpointThickness: int | float
    tieMidpointThickness: int | float
    tupletBracketThickness: int | float

    def _init(self, smufl: Smufl | None = None, auto: bool = AUTO) -> None:
        self._smufl = smufl
        self._auto = auto

    def _reprContents(self) -> list[str]:
        contents = []
        if self.font is not None:
            contents.append("in font")
            contents += self.font._reprContents()  # pylint: disable-next=W0212
            contents.append(f"auto={self._auto}")
        return contents

    def naked(self):
        # BaseObject override for __eq__ and __hash__
        return self  # pragma: no cover

    # -------
    # Parents
    # -------

    @property
    def smufl(self) -> Smufl | None:
        """Parent :class:`.smufl.Smufl` object.

        Example:

            >>> engravingDefaults.smufl  # doctest: +ELLIPSIS
            <Smufl in glyph 'uniE050' ['gClef'] ('public.default') at ...>


        """
        return self._smufl

    @smufl.setter
    def smufl(self, value: Smufl) -> None:
        if self._smufl is not None and self._smufl != value:
            raise AssertionError(
                "Smufl for EngravingDefaults object is "
                "already set and is not same as value."
            )
        self._smufl = normalizers.normalizeSmufl(value)

    @property
    def font(self) -> Font | None:
        """Parent :class:`.Font` object.

        Example:

            >>> engravingDefaults.font  # doctest: +ELLIPSIS
            <Font 'MyFont Regular' path='/path/to/MyFont.ufo' at ...>

        """
        if self._smufl is None:
            return None
        return self._smufl.font

    @property
    def glyph(self) -> Glyph | None:
        """Parent :class:`.Glyph` object.

        Example:

            >>> engravingDefaults.glyph  # doctest: +ELLIPSIS
            <Glyph 'uniE050' ['gClef'] ('public.default') at ...>

        """
        if self._smufl is None:
            return None
        return self._smufl.glyph

    @property
    def layer(self) -> Layer | None:
        """Parent :class:`.Layer` object.

        Example:

        >>> engravingDefaults.layer  # doctest: +ELLIPSIS
        <Layer 'public.default' at ...>

        """
        if self._smufl is None:
            return None
        return self._smufl.layer

    # ----
    # Auto
    # ----

    @property
    def auto(self) -> bool:
        """Whether to calculate engraving defaults automatically.

        When :obj:`True`, engraving defaults are calculated automatically from font
        measurements. When :obj:`False`, all values must be set explicitly.

        Automatically calculated values are overridden by any explicitly set defaults,
        regardless of the `auto` setting.

        Example:

            >>> engravingDefaults.auto = False
            >>> engravingDefaults.auto
            False

        """
        return self._auto

    @auto.setter
    def auto(self, value: bool) -> None:
        self._auto = value

    # ------------------
    # Dictionary methods
    # ------------------

    def clear(self) -> None:
        """Clear all engraving default settings.

        If :confval:`engravingDefaults.auto` is enabled, values will fall back to
        automatically calculated values after being cleared.

        Example::

            >>> engravingDefaults.items()
            {'arrowShaftThickness': 46, 'barlineSeparation': 72, ...}
            >>> engravingDefaults.clear()
            >>> engravingDefaults.items()
            {'arrowShaftThickness': None, 'barlineSeparation': None, ...}

        """
        if self.font is not None:
            self.font.lib.naked().pop("com.smufolib.engravingDefaults", None)

    def items(self) -> EngravingDefaultsDictReturn:
        """Return :class:`dict` of all available settings and their values.

        Example:

            >>> engravingDefaults.items()  # doctest: +ELLIPSIS
            {'arrowShaftThickness': 46, 'barlineSeparation': 72, ...}

        """
        return dict(zip(self.keys(), self.values()))

    def keys(self) -> list[str]:
        """Return sorted :class:`list` of all available settings names.

        Example:

            >>> engravingDefaults.keys()  # doctest: +ELLIPSIS
            ['arrowShaftThickness', 'barlineSeparation', ...]

        """
        return sorted(ENGRAVING_DEFAULTS_ATTRIBUTES)

    def update(
        self,
        other: EngravingDefaults | EngravingDefaultsDictInput | None = None,
        **kwargs: EngravingDefaultsInput | None,
    ) -> None:
        r"""Update settings attributes with other object or values.

        :param other: Other :class:`EngravingDefaults` or :class:`dict`
            of attribute names mapped to values. Defaults to `None`.
        :param \**kwargs: Attribute names and values to update as
         keyword arguments.

        An object may be updated in a few different ways:

            >>> engravingDefaults.items()  # doctest: +ELLIPSIS
            {'arrowShaftThickness': 46, 'barlineSeparation': 72, ...}

        # Update with :class:`dict` object.:

            >>> other = {"arrowShaftThickness": 35, "barlineSeparation": 68}
            >>> engravingDefaults.update(other)
            >>> engravingDefaults.items()  # doctest: +ELLIPSIS
            {'arrowShaftThickness': 35, 'barlineSeparation': 68, ...}

        Update with other :class:`EngravingDefaults` object.:

            >>> other = otherFont.smufl.engravingDefaults
            >>> other.items()
            {'arrowShaftThickness': 52, 'barlineSeparation': 75, ...}
            >>> engravingDefaults.update(other)
            >>> engravingDefaults.items()  # doctest: +ELLIPSIS
            {'arrowShaftThickness': 52, 'barlineSeparation': 75, ...}

        Update with keyword arguments:

            >>> engravingDefaults.update(arrowShaftThickness=46, barlineSeparation=72)
            >>> engravingDefaults.items()  # doctest: +ELLIPSIS
            {'arrowShaftThickness': 46, 'barlineSeparation': 72, ...}

        """
        normalizedOther = self._mergeOthers(other, kwargs)
        if not normalizedOther or self.font is None:
            return

        if "com.smufolib.engravingDefaults" not in self.font.lib:
            self.font.lib["com.smufolib.engravingDefaults"] = {}

        for key, value in normalizedOther.items():
            if key not in self.keys():
                raise AttributeError(
                    error.generateErrorMessage(
                        "attributeError", objectName=type(self).__name__, attribute=key
                    )
                )

            normalizedValue = normalizers.normalizeEngravingDefaultsAttr(key, value)
            if value is None:
                continue
            if (
                self._smufl is not None
                and isinstance(normalizedValue, (int, float))
                and self.spaces
            ):
                normalizedValue = self._smufl.toUnits(normalizedValue)
            self.font.lib["com.smufolib.engravingDefaults"][key] = normalizedValue

    def _mergeOthers(
        self,
        other: EngravingDefaults | EngravingDefaultsDictInput | None,
        kwargs: EngravingDefaultsDictInput,
    ) -> EngravingDefaultsDictInput | None:
        error.validateType(
            other,
            (EngravingDefaults, dict, type(None)),
            f"{self.__class__.__name__}.update()",
        )

        base: dict
        if isinstance(other, EngravingDefaults):
            base = other.items()
        elif other is None:
            if not kwargs:
                return None
            base = {}
        else:
            base = other

        base |= kwargs
        return base

    def values(self) -> list[EngravingDefaultsReturn]:
        """Return a :class:`list` of all available settings values.

        Order corresponds to :meth:`keys`.

        Example:

            >>> engravingDefaults.values()  # doctest: +ELLIPSIS
            [46, 72, ...]

        """
        return [getattr(self, k) for k in self.keys()]

    # ----------
    # Attributes
    # ----------

    @property
    def textFontFamily(self) -> tuple[str] | None:
        """Preferred text font families for pairing with this music font"""
        return cast(tuple[str] | None, self._getValue("textFontFamily"))

    @textFontFamily.setter
    def textFontFamily(self, value: CollectionType[str]) -> None:
        self._setValue("textFontFamily", value)

    # All other properties declared dynamically below

    def _getValue(self, name: str) -> EngravingDefaultsReturn | None:
        # Common settings property getter.
        if not self._libDict and not self._auto:
            return None

        value = self._libDict.get(name, None)
        if name == "textFontFamily":
            value = self._libDict.get(name, ())

        if self.font and value is None and self._auto:
            glyphName = ENGRAVING_DEFAULTS_MAPPING[name]["glyph"]
            try:
                glyph = self.font[glyphName]
            except KeyError:
                return None
            rulerName = ENGRAVING_DEFAULTS_MAPPING[name]["ruler"]
            ruler: Callable[["Glyph"], int | float | None] = DISPATCHER[rulerName]
            value = ruler(glyph)

        elif self.spaces and isinstance(value, (int, float)) and self.font is not None:
            return self.font.smufl.toSpaces(value)
        return value

    def _setValue(self, name: str, value: EngravingDefaultsInput | None) -> None:
        # Common settings property setter.
        # Keeps dynamic dict of settings in font.lib().

        if self.font is None:
            return

        normalizedValue = normalizers.normalizeEngravingDefaultsAttr(name, value)
        libDict = self._libDict

        if not normalizedValue:
            libDict.pop(name, None)

            if not libDict:
                self.font.lib.pop("com.smufolib.engravingDefaults", None)
            else:
                self._libDict = libDict

        else:
            if self.spaces and isinstance(normalizedValue, (int, float)):
                normalizedValue = self.font.smufl.toUnits(normalizedValue)

            self._libDict[name] = normalizedValue
            self._libDict = libDict

    @property
    def _libDict(self) -> EngravingDefaultsDictReturn:
        # Dynamic dict in font.lib.naked().
        if self.font is None:
            return {}
        lib = self.font.lib.naked()
        return lib.setdefault("com.smufolib.engravingDefaults", {})

    @_libDict.setter
    def _libDict(self, value: EngravingDefaultsDictInput) -> None:
        if self.font is not None:
            self.font.lib.naked()["com.smufolib.engravingDefaults"] = value

    # -----------------------------
    # Normalization and Measurement
    # -----------------------------

    def round(self) -> None:
        """Round font units to integers.

        Method applies to the following attributes:

        * :attr:`arrowShaftThickness`
        * :attr:`barlineSeparation`
        * :attr:`beamSpacing`
        * :attr:`beamThickness`
        * :attr:`bracketThickness`
        * :attr:`dashedBarlineDashLength`
        * :attr:`dashedBarlineGapLength`
        * :attr:`dashedBarlineThickness`
        * :attr:`hairpinThickness`
        * :attr:`hBarThickness`
        * :attr:`legerLineExtension`
        * :attr:`legerLineThickness`
        * :attr:`lyricLineThickness`
        * :attr:`octaveLineThickness`
        * :attr:`pedalLineThickness`
        * :attr:`repeatBarlineDotSeparation`
        * :attr:`repeatEndingLineThickness`
        * :attr:`slurEndpointThickness`
        * :attr:`slurMidpointThickness`
        * :attr:`staffLineThickness`
        * :attr:`stemThickness`
        * :attr:`subBracketThickness`
        * :attr:`textEnclosureThickness`
        * :attr:`thickBarlineThickness`
        * :attr:`thinBarlineThickness`
        * :attr:`thinThickBarlineSeparation`
        * :attr:`tieEndpointThickness`
        * :attr:`tieMidpointThickness`
        * :attr:`tupletBracketThickness`

        If :attr:`spaces` is :obj:`True`, values are left unchanged.

        Example:

            >>> engravingDefaults.spaces = True
            >>> engravingDefaults.stemThickness = 0.12
            >>> engravingDefaults.round()
            >>> engravingDefaults.stemThickness
            0.12

            >>> engravingDefaults.spaces = False
            >>> engravingDefaults.stemThickness = 30.5
            >>> engravingDefaults.round()
            >>> engravingDefaults.stemThickness
            31

        """
        if self.spaces:
            return

        for key in self.keys():
            attribute = getattr(self, key)
            if not isinstance(attribute, (int, float)):
                continue
            value = normalizers.normalizeVisualRounding(attribute)
            setattr(self, key, value)

    @property
    def spaces(self) -> bool:
        """Whether to set state of measurement to staff spaces.

        Example:

            engravingDefaults.stemThickness = 25
            >>> engravingDefaults.spaces = True
            >>> engravingDefaults.stemThickness
            0.1
            >>> engravingDefaults.spaces = False
            >>> engravingDefaults.stemThickness
            25


        """
        if self._smufl is None:
            return False
        return self._smufl.spaces

    @spaces.setter
    def spaces(self, value: bool) -> None:
        if self._smufl is None:
            return
        self._smufl.spaces = value

    # ------------------------
    # Override from BaseObject
    # ------------------------

    def raiseNotImplementedError(self):
        """This exception needs to be raised frequently by
        the base classes. So, it's here for convenience.
        """
        raise NotImplementedError(  # pragma: no cover
            error.generateErrorMessage(
                "notImplementedError", objectName=self.__class__.__name__
            )
        )


def _makeProperty(attr: str, doc: str):
    # textFontFamily declared explicitly above

    def getter(self) -> int | float | None:
        return self._getValue(attr)

    def setter(self, value: int | float | None) -> None:
        self._setValue(attr, value)

    return property(getter, setter, doc=doc)


for name, description in ENGRAVING_DEFAULTS_ATTRIBUTES.items():
    if name != "textFontFamily":
        setattr(EngravingDefaults, name, _makeProperty(name, description))



# pylint: disable=C0114
from __future__ import annotations
from typing import Any
from copy import deepcopy
from argparse import (
    ArgumentParser,
    HelpFormatter,
    RawDescriptionHelpFormatter,
    RawTextHelpFormatter,
    ArgumentDefaultsHelpFormatter,
    MetavarTypeHelpFormatter,
)
import json

from smufolib import config
from smufolib.objects.font import Font
from smufolib.request import Request
from smufolib.utils import converters, error

CONFIG = config.load()

# pylint: disable=C0103
# fmt: off

#: Available arguments and their settings.
CLI_ARGUMENTS: dict[str, dict[str, Any]] = {
    "attributes": {
        "nargs": "+",
        "help": "attribute names to include in processing"
    },
    "classesData": {
        "type": Request,
        "help": "path to classes metadata file"
    },
    "clear": {
        "action": "store_true",
        "help": "erase preexisting objects on execution"
    },
    "color": {
        "nargs": 4,
        "type": converters.toNumber,
        "help": "list of RGBA color values"
    },
    "colors": {
        "type": json.loads,
        "help": "keys mapped to RGBA color arrays as JSON string"
    },
    "exclude": {
        "nargs": "+",
        "help": "objects to exclude from processing"
    },
    "font": {
        "type": Font,
        "help": "path to UFO file"
    },
    "fontData": {
        "type": Request,
        "help": "path to font metadata file"
    },
    "glyphnamesData": {
        "type": Request,
        "help": "path to glyphnames metadata file"
    },
    "include": {
        "nargs": "+",
        "help": "objects to include in processing"
    },
    "includeOptionals": {
        "action": "store_true",
        "help": "include optional glyphs"
    },
    "mark": {
        "action": "store_true",
        "help": "apply defined color values to objects"
    },
    "overwrite": {
        "action": "store_true",
        "help": "overwrite preexisting values"
    },
    "rangesData": {
        "type": Request,
        "help": "path to ranges metadata file"
    },
    "sourcePath": {
        "type": Request,
        "help": "path to source file or directory"
    },
    "spaces": {
        "action": "store_true",
        "help": "set unit of measurement to staff spaces"
    },
    "targetPath": {
        "help": "path to target file or directory"
    },
    "verbose": {
        "action": "store_true",
        "help": "make output verbose"
    }
}

# fmt: on


def commonParser(
    *args: str,
    addHelp: bool = True,
    description: str | None = None,
    customHelpers: dict[str, str] | None = None,
    **kwargs: Any,
) -> ArgumentParser:
    r"""Provide generic command-line arguments and options.

    See the :ref:`Available Options` for details.

    :param \*args: Required positional arguments to assign.
    :param addHelp: Add help message. Should be :obj:`False` when
        function is parent and otherwise :obj:`True`. Defaults
        to :obj:`True`.
    :param description: Program description when used directly. Defaults
        to :obj:`None`
    :param customHelpers: Arguments mapped to custom help strings to
        override the default. Defaults to :obj:`None`
    :param \**kwargs: Options and their default values to assign.

    Example:

        >>> from smufolib import commonParser
        >>> parser = commonParser(
        ...     "font", "clear", includeOptionals=False,
        ...     description="My SMuFL utility", addHelp=True
        ...     )

    """

    parser = ArgumentParser(add_help=addHelp, description=description)
    localArguments = deepcopy(CLI_ARGUMENTS)

    def addArgument(
        arg: str, flags: tuple[str, ...], customHelpers: dict[str, str] | None
    ) -> None:
        # Add argument to parser.
        if customHelpers and arg in customHelpers:
            localArguments[arg]["help"] = customHelpers[arg]
        parser.add_argument(*flags, **localArguments[arg])

    def generateFlags(argument: str) -> tuple[str, str]:
        # Generates tuple of option flags.
        shortFlags = CONFIG["cli.shortFlags"]
        longFlag = f"--{converters.toKebab(argument)}"
        return (shortFlags[argument], longFlag)

    for arg in args:
        flags: tuple[str, ...] = generateFlags(arg)
        if localArguments[arg].get("action") != "store_true":
            flags = (arg,)
            localArguments[arg]["metavar"] = converters.toKebab(arg)
        addArgument(arg, flags, customHelpers)

    for key, value in kwargs.items():
        if key in args:
            raise ValueError(error.generateErrorMessage("argumentConflict", key=key))

        flags = generateFlags(key)
        localArguments[key]["dest"] = key
        if value is not None:
            localArguments[key]["default"] = value
        addArgument(key, flags, customHelpers)

    return parser


def createHelpFormatter(formatters: str | tuple[str, ...]) -> type[HelpFormatter]:
    """Create child class of multiple help formatters.

    The returned :class:`argparse.HelpFormatter` class can be passed to the
    `formatter_class` parameter of :class:`argparse.ArgumentParser` to
    combine different formatters, despite the parameter only taking a
    single class as argument.

    :param formatters: Name of the formatter class or tuple of class
        names to be enabled as :class:`str` corresponding to
        either :class:`~argparse.RawDescriptionHelpFormatter`,
        :class:`~argparse.RawTextHelpFormatter`,
        :class:`~argparse.ArgumentDefaultsHelpFormatter`,
        :class:`~argparse.RawDescriptionHelpFormatter` or
        :class:`~argparse.MetavarTypeHelpFormatter`, or a :class:`tuple`
        of class names.
    :raises TypeError: If `formatters` is not an accepted type.
    :raises ValueError: If any `formatters` item is not recognized.

    Example:

        >>> import argparse
        >>> from smufolib import cli
        >>> formatter = cli.createHelpFormatter(
        ...    ("RawTextHelpFormatter", "ArgumentDefaultsHelpFormatter")
        ... )
        >>> parser = argparse.ArgumentParser(
        ...     formatter_class=formatter,
        ...     description="Process SMuFL metadata"
        ... )

    """

    baseFormatters = {
        "HelpFormatter": HelpFormatter,
        "RawDescriptionHelpFormatter": RawDescriptionHelpFormatter,
        "RawTextHelpFormatter": RawTextHelpFormatter,
        "ArgumentDefaultsHelpFormatter": ArgumentDefaultsHelpFormatter,
        "MetavarTypeHelpFormatter": MetavarTypeHelpFormatter,
    }

    error.validateType(formatters, (str, tuple), "formatters")
    if isinstance(formatters, str):
        formatters = (formatters,)

    try:
        employed = tuple(baseFormatters[v] for v in formatters)
    except KeyError as exc:
        raise ValueError(
            error.suggestValue(
                # Casting exc to str to please mypy.
                str(exc),
                list(baseFormatters),
                "formatters",
                items=True,
            )
        ) from exc

    return type("HelpFormatter", employed, {})



#!/usr/bin/env python3
"""Import SMuFL identification attributes.

This script imports :class:`.Smufl` identification attribute values for all glyphs in a
SMufoLib :class:`~smufolib.objects.font.Font` from metadata files. More specifically, it
automatically sets the values for :attr:`.Smufl.name`, :attr:`.Smufl.description` and/or
:attr:`.Smufl.classes`, based on the official metadata files of :smufl:`SMuFL
<specification/smufl-metadata.html>` and `Bravura
<https://github.com/steinbergmedia/bravura#bravura-music-font>`_, or any other
compatible sources.

The script supports importing optional glyphs and controlling whether existing data is
overwritten. Glyphs are skipped if they are not valid SMuFL members or if their
metadata cannot be found.

The script requires SMufoLib to be installed in its execution environment. It can be
used from the command line or as a Python module. See the :ref:`import-id-cli` and
:ref:`import-id-python` sections below for usage details.

"""

from __future__ import annotations
from typing import Any
import argparse
import collections
from contextlib import nullcontext
from pathlib import Path
import time

from tqdm import tqdm

from smufolib import (
    Font,
    Request,
    cli,
    config,
    converters,
    error,
    stdUtils,
)
from smufolib.utils.scriptUtils import normalizeFont as _normalizeFont
from smufolib.utils.scriptUtils import normalizeJsonDict as _normalizeJsonDict
from smufolib.utils.scriptUtils import normalizeRequest as _normalizeRequest

# Type aliases
JsonDict = dict[str, Any]
GlyphMap = dict[int, str | list[str]]
AttributesMap = dict[str, GlyphMap]

# Configuration
CONFIG = config.load()

# Parameter defaults
CLASSES_DATA = Request(
    CONFIG["metadata.paths"]["classes"], CONFIG["metadata.fallbacks"]["classes"]
)
GLYPHNAMES_DATA = Request(
    CONFIG["metadata.paths"]["glyphnames"], CONFIG["metadata.fallbacks"]["glyphnames"]
)
FONT_DATA = Request(
    CONFIG["metadata.paths"]["font"],
    CONFIG["metadata.fallbacks"]["font"],
)
ATTRIBUTES = ("name", "classes", "description")
INCLUDE_OPTIONALS = False
OVERWRITE = False
VERBOSE = False

# pylint: disable=R0913, R0914, C0103


def importID(
    font: Font | Path | str,
    attributes: str | tuple[str, ...] = "*",
    classesData: Request | Path | str = CLASSES_DATA,
    glyphnamesData: Request | Path | str = GLYPHNAMES_DATA,
    fontData: Request | Path | str = FONT_DATA,
    includeOptionals: bool = INCLUDE_OPTIONALS,
    overwrite: bool = OVERWRITE,
    verbose: bool = VERBOSE,
) -> None:
    """Import SMuFL identification attributes (Python API).

    Optional glyphs can be included by setting `includeOptionals` to :obj:`True`. When
    enabled, stylistic alternates and ligatures must be named with reference to their
    base glyph (see :ref:`this note <about-glyph-naming>` for more details).

    If `overwrite` is :obj:`False`, glyphs with existing attribute values are skipped.
    Glyphs are also skipped if they are not SMuFL members or if lookup fails (e.g., due
    to missing codepoints or unencoded glyphs in the metadata).

    :param font: Object or path to target :class:`.Font`.
    :param attributes: ID attributes to be set. Value can be either ``"*"`` (all),
        ``"name"``, ``"classes"``, ``"description"`` or :class:`tuple` of several.
        Defaults to ``"*"``.
    :param classesData: Request for or path to classes metadata file. Defaults to
        :class:`smufolib.request.Request` passing :confval:`metadata.paths.classes` and
        :confval:`metadata.fallbacks.classes`.
    :param glyphnamesData: Request for or path to glyphnames metadata file. Defaults to
        :class:`smufolib.request.Request` passing :confval:`metadata.paths.glyphnames`
        and :confval:`metadata.fallbacks.glyphnames`.
    :param fontData: Request for or path to reference font metadata file. Defaults to
        :class:`smufolib.request.Request` passing :confval:`metadata.paths.font` and
        :confval:`metadata.fallbacks.font`.
    :param includeOptionals: Include optional glyphs. Defaults to :obj:`False`.
    :param overwrite: Overwrite preexisting values. Defaults to :obj:`False`.
    :param verbose: Make output verbose. Defaults to :obj:`False`.
    :raises TypeError: If any parameter value is not the expected type.
    :raises ValueError: If `attributes` value is not a valid ID attribute.

    """
    print("Starting...")

    font = _normalizeFont(font)
    ticks = len(font) * 2 + 2
    with tqdm(total=ticks) if not verbose else nullcontext() as progressBar:
        attributes = _normalizeAttributes(attributes)
        classesDataJson = _normalizeJsonDict(_normalizeRequest(classesData).json())
        glyphnamesDataJson = _normalizeJsonDict(
            _normalizeRequest(glyphnamesData).json()
        )
        fontDataJson = _normalizeJsonDict(_normalizeRequest(fontData).json())
        fontDataJson = fontDataJson.get("optionalGlyphs", {}) | fontDataJson.get(
            "ligatures", {}
        )

        if progressBar:
            progressBar.update(1)
            time.sleep(0.0001)

        stdUtils.verbosePrint("\nBuilding attributesMap...", verbose)
        attributesMap = _buildAttributesMap(
            font=font,
            attributes=attributes,
            includeOptionals=includeOptionals,
            classesDataJson=classesDataJson,
            glyphnamesDataJson=glyphnamesDataJson,
            fontDataJson=fontDataJson,
            progressBar=progressBar,
        )

        for glyph in font:
            if progressBar:
                progressBar.update(1)
                time.sleep(0.0001)

            codepoint = glyph.unicode
            if not codepoint:
                stdUtils.verbosePrint(
                    f"\nSkipping unencoded glyph: '{glyph.name}'", verbose
                )
                continue

            if not glyph.smufl.isMember:
                stdUtils.verbosePrint(
                    f"\nSkipping non-SMuFL glyph: '{glyph.name}'", verbose
                )
                continue

            if glyph.smufl.isOptional and not includeOptionals:
                stdUtils.verbosePrint(
                    f"\nSkipping optional glyph: '{glyph.name}'", verbose
                )
                continue

            stdUtils.verbosePrint(
                f"\nSetting attributes for glyph '{glyph.name}':", verbose
            )
            for attribute in attributes:
                if codepoint not in attributesMap[attribute]:
                    stdUtils.verbosePrint(f"\t'{attribute}': not found", verbose)
                    continue

                if getattr(glyph.smufl, attribute) and not overwrite:
                    stdUtils.verbosePrint(f"\t'{attribute}': preset", verbose)
                    continue

                setattr(glyph.smufl, attribute, attributesMap[attribute][codepoint])
                stdUtils.verbosePrint(f"\t'{attribute}': set", verbose)

        stdUtils.verbosePrint("\nSaving font...", verbose)
        font.save()
        if progressBar:
            progressBar.update(1)
            time.sleep(0.0001)

    print("\nDone.")


def main() -> None:
    """Command line entry point."""
    args = _parseArgs()
    importID(
        args.font,
        args.attributes,
        classesData=args.classesData,
        glyphnamesData=args.glyphnamesData,
        fontData=args.fontData,
        includeOptionals=args.includeOptionals,
        overwrite=args.overwrite,
        verbose=args.verbose,
    )


def _normalizeAttributes(value: str | tuple[str, ...] | list[str]) -> tuple[str, ...]:
    # Normalize values in the ``attributes`` parameter.

    value = ATTRIBUTES if value in ("*", ["*"]) else value
    value = (value,) if isinstance(value, str) else value
    error.validateType(value, (str, tuple, list), "attributes")
    for val in value:
        if val not in ATTRIBUTES:
            error.suggestValue(val, ATTRIBUTES, "attributes", items=True)
    return tuple(value)


def _buildAttributesMap(
    font: Font,
    attributes: str | tuple[str, ...],
    includeOptionals: bool,
    classesDataJson: JsonDict,
    glyphnamesDataJson: JsonDict,
    fontDataJson: JsonDict,
    progressBar: tqdm,
) -> AttributesMap:
    # Build dictionary of ID attributes mapped to glyph maps.

    def _buildGlyphMap(attribute: str, metadata: JsonDict) -> GlyphMap:
        # Build dictionary of codepoints mapped to attribute values.

        # Descriptions for ligatures are not included in
        # the 'OptionalGlyphs' section of bravura_metadata.json, and
        # lookups in both sections are therefore needed.

        glyphMap: GlyphMap = {}
        for name, data in metadata.items():
            codepoint = converters.toDecimal(data["codepoint"])
            if attribute == "name":
                glyphMap[codepoint] = name
            else:
                glyphMap[codepoint] = data.get(attribute, None)

        return glyphMap

    def _buildClassMap(
        font: Font, classesDataJson: JsonDict, glyphnamesDataJson: JsonDict
    ) -> GlyphMap:
        # Build glyph map for 'classes'.

        # GlyphnamesData has a number of unused codepoints and
        # codepoints for text based fonts (58 as of this writing) which
        # should be skipped.

        # Invert classesData dictionary to glyph names mapped to classes.
        classMap: GlyphMap = {}
        classes = collections.defaultdict(list)
        for clas, glyphs in classesDataJson.items():
            for glyph in glyphs:
                name = converters.toUniName(glyphnamesDataJson[glyph]["codepoint"])
                classes[name].append(clas)

        # Loop through basenames to get classes for optional glyphs and
        # index on codepoints.
        for glyph in font:
            if progressBar:
                progressBar.update(1)
                time.sleep(0.0001)

            baseGlyph = glyph.smufl.base
            if not glyph.smufl.codepoint or not glyph.smufl.isMember:
                continue

            if baseGlyph.name not in classes:
                continue

            classMap[glyph.unicode] = classes[baseGlyph.name]

        return classMap

    attributesMap: AttributesMap = {}
    for attribute in attributes:
        if attribute == "classes":
            attributesMap[attribute] = _buildClassMap(
                font=font,
                classesDataJson=classesDataJson,
                glyphnamesDataJson=glyphnamesDataJson,
            )
            if includeOptionals:
                attributesMap[attribute] |= _buildGlyphMap(
                    attribute=attribute, metadata=fontDataJson
                )
        else:
            attributesMap[attribute] = _buildGlyphMap(
                attribute=attribute, metadata=glyphnamesDataJson
            )
            if includeOptionals:
                attributesMap[attribute] |= _buildGlyphMap(
                    attribute=attribute, metadata=fontDataJson
                )

    return attributesMap


def _parseArgs() -> argparse.Namespace:
    # Parse command line arguments and options.
    parser = cli.commonParser(
        "font",
        description=stdUtils.getSummary(importID.__doc__),
        attributes="*",
        classesData=CLASSES_DATA,
        glyphnamesData=GLYPHNAMES_DATA,
        fontData=FONT_DATA,
        includeOptionals=INCLUDE_OPTIONALS,
        overwrite=OVERWRITE,
        verbose=VERBOSE,
    )
    return parser.parse_args()


if __name__ == "__main__":
    main()



"""Utility functions for converting values between various formats.

These functions focus on conversion different string and number formats
related to measurements, Unicode codepoints and letter case.

To import the module:

    >>> from smufolib import converters

"""

from __future__ import annotations
import re

from smufolib.utils import error, stdUtils


# pylint: disable=C0103
def convertMeasurement(
    measurement: int | float | str,
    targetUnit: str,
    unitsPerEm: int | float,
    rounded: bool = False,
) -> int | float:
    """Convert between units of measurement based on UPM size.

    :param measurement: The value to convert.
    :param targetUnit: Unit to convert to.
    :param unitsPerEm: The UPM value to base conversion on.
    :param rounded: Whether to round result to nearest integer.
    :raises TypeError: If `measurement` is not an accepted type.
    :raises ValueError: If `targetUnit` is neither ``"spaces"``
     nor ``"units"``.

    Example:

        >>> converters.convertMeasurement(0.795, "units", 1000, rounded=False)
        198.75
        >>> converters.convertMeasurement(0.795, "units", 1000, rounded=True)
        199
        >>> converters.convertMeasurement("198.75", "spaces", 1000, rounded=True)
        0.795

    """
    error.validateType(measurement, (int, float, str), "measurement")
    error.suggestValue(targetUnit, ("spaces", "units"), "targetUnit")

    space = unitsPerEm / 4
    measurement = float(measurement)
    if targetUnit == "spaces":
        measurement /= space
        return round(measurement, 3)

    measurement *= space
    return round(measurement) if rounded else measurement


def toDecimal(unicodeString: str) -> int:
    """Convert formatted unicode or uni name to decimal codepoint.

    Function accepts any hexadecimal string within the Unicode range
    (U+0000 -- U+10FFFF) prefixed by ``"u"``, ``"U+"`` or ``"uni"``.

    :param unicodeString: The value to convert.
    :raises TypeError: If `unicodeString` is not the accepted type.
    :raises ValueError: If `unicodeString` is not a valid formatted
        unicode codepoint or outside the unicode range
        (U+0000 -- U+10FFFF).

    Example:

        >>> converters.toDecimal("uniE000")
        57344
        >>> converters.toDecimal("u1D100")
        119040
        >>> converters.toDecimal("U+0A00")
        2560

    """
    error.validateType(unicodeString, str, "unicodeString")
    unicodeHex = _findUnicodeHex(unicodeString)
    if not unicodeHex:
        raise ValueError(
            error.generateErrorMessage("invalidFormat", objectName="unicodeString")
        )

    decimal = int(toNumber(unicodeHex))  # Casting to int to please mypy.
    if _isInUnicodeRange(decimal):
        return decimal

    raise ValueError(
        error.generateErrorMessage("unicodeOutOfRange", objectName="unicodeString")
    )


def toUniHex(codepoint: int) -> str:
    """Convert decimal codepoint to formatted Unicode hex.

    :param codepoint: The decimal value to convert.
    :raises TypeError: If `codepoint` is not the accepted type.
    :raises ValueError: If `codepoint` is outside the Unicode range
        (U+0000 -- U+10FFFF).

    Example:

        >>> converters.toUniHex(2560)
        'U+0A00'

    """
    error.validateType(codepoint, int, "codepoint")

    if _isInUnicodeRange(codepoint):
        return f"U+{str(hex(codepoint).upper()[2:]).zfill(4)}"
    raise ValueError(
        error.generateErrorMessage("unicodeOutOfRange", objectName="codepoint")
    )


def toUniName(value: str | int, short: bool = False) -> str:
    """Convert formatted Unicode hex or decimal to ``"uni"`` name.

    Function accepts any integer or prefixed hexadecimal string
    within the Unicode range (U+0000 -- U+10FFFF).

    :param value: The value to convert.
    :param short: Whether to return name with single ``"u"`` prefix.
    :raises TypeError: If `value` is not an accepted type.
    :raises ValueError: If `value` is outside the Unicode range
        (U+0000 -- U+10FFFF) or not a valid formatted string.

    Example:

        >>> converters.toUniName("U+E000")
        'uniE000'
        >>> converters.toUniName('U+E000', short=True)
        'uE000'
        >>> converters.toUniName(57344)
        'uniE000'

    """
    prefix = "u" if short else "uni"

    error.validateType(value, (str, int), "value")
    if isinstance(value, str):
        unicodeHex = _findUnicodeHex(value)
        if unicodeHex is None:
            raise ValueError(
                error.generateErrorMessage("invalidFormat", objectName="value")
            )
    else:
        unicodeHex = "0x" + format(value, "X").zfill(4)

    if not _isInUnicodeRange(int(unicodeHex, 16)):
        raise ValueError(
            error.generateErrorMessage("unicodeOutOfRange", objectName="value")
        )

    return prefix + unicodeHex[2:].upper()


def toKebab(camelCaseString: str) -> str:
    """Convert camelCase to kebab-case.

    :param camelCaseString: The string to convert.

    Example:

        >>> converters.toKebab("camelCase")
        'camel-case'

    """
    return re.sub(r"(?<!^)(?=[A-Z]{1})", "-", camelCaseString).lower()


def toNumber(numericString: str) -> int | float:
    """Convert numeric string to number.

    :param numericString: The value to convert.
    :raises ValueError: If `numericString` is not a valid int or float.

    Example:

        >>> converters.toNumber("57344")
        57344
        >>> converters.toNumber("57.344")
        57.344
        >>> converters.toNumber("E000")
        57344

    """
    error.validateType(numericString, str, "numericString")
    try:
        if stdUtils.isFloat(numericString):
            return float(numericString)
        if numericString.isnumeric():
            return int(numericString)
        return int(numericString, 16)
    except ValueError as exc:
        raise ValueError(
            error.generateErrorMessage("numericValue", objectName="numericString")
        ) from exc


def toIntIfWhole(number: int | float) -> int | float:
    """Convert float without a fractional part to integer.

    If `number` is already an :class:`int` or the :class:`float` has a
    fractional part, it is left unchanged.

    :param number: The value to convert.

    Example:

        >>> converters.toIntIfWhole(34.0)
        34
        >>> converters.toIntIfWhole(34.1)
        34.1

    """

    if isinstance(number, float) and number.is_integer():
        return int(number)
    return number


def _findUnicodeHex(unicodeString: str) -> str | None:
    # Find the hex in various unicode value configurations.
    # If found, add prefix and return value or None.
    error.validateType(unicodeString, str, "unicodeString")
    pattern = r"((?<=^u)|(?<=^u\+)|(?<=^uni))([a-f]|[0-9]){4,}"
    result = re.search(pattern, unicodeString, flags=re.IGNORECASE)
    if not result:
        return None
    return f"0x{result.group(0)}"


def _isInUnicodeRange(codepoint: int) -> bool:
    # Check if number is within Unicode range
    if 0x0 <= codepoint <= 0x10FFFF:
        return True
    return False



"""Utility functions for error message generation and type validation.

This module provides functions to generate and validate error messages,
check types, and suggest corrections for invalid values. It includes
a dictionary of error message templates to ensure streamlined and
consistent error reporting.

To import the module:

    >>> from smufolib import error

"""

from __future__ import annotations
from types import UnionType
from typing import Any, get_args
import difflib

# pylint: disable=C0103

#: Dictionary of error message templates.
ERROR_TEMPLATES: dict[str, str] = {
    "alphanumericValue": "The value for {objectName!r} must be alphanumeric",
    "alphanumericValueItems": "Value items for {objectName!r} must be alphanumeric",
    "argumentConflict": "The option '{key}' is already added as positional argument or flag",
    "attributeError": "{objectName!r} has no attribute {attribute!r}",
    "contextualAttributeError": "The attribute {attribute!r} is not available when {context}",
    "contextualSetAttributeError": "Cannot set attribute {attribute!r} when {context}",
    "contextualTypeError": "Expected {objectName!r} to be of type {validTypes} when {context}, but got {valueType}",
    "contextualItemsTypeError": "Items in {objectName!r} must be {validTypes} when {context}, not {valueType}",
    "deprecated": "{objectName!r} is deprecated and will be removed in the next version of SMufoLib (after {version})",
    "deprecatedReplacement": "Use {replacement!r} instead",
    "duplicateFlags": "Arguments {argument1!r} and {argument2!r} have duplicate short flag: {flag}",
    "duplicateAttributeValue": "The value {value!r} for {attribute!r} is already assigned to another {objectName} instance: {conflictingInstance!r}",
    "duplicateItems": "Items in {objectName!r} cannot be duplicates",
    "emptyValue": "The value for {objectName!r} cannot be empty",
    "emptyValueItems": "Value items for {objectName!r} cannot be empty",
    "fileNotFound": "The file or directory for {objectName!r} does not exist",
    "invalidFormat": "The value for {objectName!r} is not correctly formatted",
    "invalidInitialCharacter": "The value for {objectName!r} must start with a lowercase letter or number",
    "invalidInitialItemsCharacter": "Value items for {objectName!r} must start with a lowercase letter or number",
    "itemsTypeError": "Items in {objectName!r} must be {validTypes}, not {valueType}",
    "itemsValueError": "Invalid value for item in {objectName!r}: {value!r}",
    "missingExtension": "The value for {objectName!r} must have a {extension!r} extension",
    "missingDependencyError": "Cannot set {objectName!r} because {dependency!r} is None",
    "missingValue": "Required values for {objectName!r} are missing",
    "nonIncreasingRange": "The values in {objectName!r} must form an increasing range",
    "notImplementedError": "The {objectName!r} subclass does not implement this method",
    "numericValue": "The value for {objectName!r} must be numeric",
    "recommendScript": "Consider running the script {scriptName!r} before the current process",
    "serializationError": "Error serializing JSON data or writing to the file",
    "singleItem": "{objectName!r} must contain a value pair",
    "suggestion": "Did you mean {suggestion!r}?",
    "typeError": "Expected {objectName!r} to be of type {validTypes}, but got {valueType}",
    "unicodeOutOfRange": "The value for {objectName!r} is outside the Unicode range (U+0000 -- U+10FFFF)",
    "urlError": "Could not connect to URL: {url!r}",
    "valueError": "Invalid value for {objectName!r}: {value!r}",
    "valueTooHigh": "The value for {objectName!r} must be {value!r} or lower",
    "valueTooLow": "The value for {objectName!r} must be {value!r} or higher",
}


class URLWarning(Warning):
    """URL connection failure warning."""


def generateErrorMessage(
    *templateNames: str, string: str | None = None, **kwargs
) -> str:
    r"""Generate an error message from a template and keyword arguments.

    The `templateNames` and `string` will be concatenated in order.

    :param \*templateNames: The error message template string, which
        should contain placeholders for keyword arguments.
    :param string: An additional message string to include.
        Defaults to :obj:`None`
    :param \**kwargs: Arbitrary keyword arguments corresponding to the
        placeholders in the template string.
    :raises KeyError: If a placeholder in the template does not have a
        corresponding keyword argument.

    Examples:

        >>> error.generateErrorMessage("alphanumericValue", objectName="unicode")
        "The value for 'unicode' must be alphanumeric"

        >>> error.generateErrorMessage(
        ...     "typeError", objectName="index", validTypes="int", valueType="str"
        ...     )
        "Expected 'index' to be of type int, but got str"

        >>> error.generateErrorMessage(
        ...     "urlError", string="Please try again", url="some/url.com"
        ...     )
        "Could not connect to URL: 'some/url.com'. Please try again"

    """
    messages = [ERROR_TEMPLATES[n].format(**kwargs) for n in templateNames]
    if string:
        messages.append(string)
    return ". ".join(messages)


def generateTypeError(
    value: Any,
    validTypes: type | tuple[type, ...],
    objectName: str,
    context: str | None = None,
    items: bool = False,
) -> str:
    """Generate a :class:`TypeError` message.

    This function generates an error message based on the number of
    valid types. By default, the message is generated from
    :obj:`ERROR_TEMPLATES`:``"typeError"``.

    If `context` in not :obj:`None` (or empty) and `items`
    is :obj:`False`, :obj:`ERROR_TEMPLATES`:``"contextualTypeError"``
    is used. If `items` is :obj:`True` and `context` is :obj:`None`,
    :obj:`ERROR_TEMPLATES`:``"itemsTypeError"`` is used. If both
    `context` is not :obj:`None` and `items` is :obj:`True`,
    :obj:`ERROR_TEMPLATES`:``"contextualItemsTypeError"`` is used.

    :param value: The value to be validated.
    :param validTypes: A :class:`tuple` or :class:`list` of valid types.
    :param objectName: The name of the object being validated.
    :param context: Additional substring for contextual type errors.
        Defaults to :obj:`None`
    :param items: Whether to use items-specific template. Defaults
        to :obj:`False`.
    :raises TypeError: If `value` does not match any of the valid
        types.
    :raises ValueError: If `items` is :obj:`True` and `value` is not
        an iterable.

    Examples:

        >>> error.generateTypeError(123, (str,), "path")
        "Expected 'path' to be of type str, but got int"

        >>> from pathlib import Path
        >>> error.generateTypeError(123, (str, Path), "path")
        "Expected 'path' to be of type str or Path, but got int"

        >>> from smufolib import Request
        >>> error.generateTypeError(123, (str, Path, Request), "path")
        "Expected 'path' to be of type str, Path or Request, but got int"

    """
    typeNames = _listTypes(validTypes)
    valueType = type(value).__name__

    kwargs = {
        "validTypes": typeNames,
        "objectName": objectName,
        "valueType": valueType,
    }

    if context:
        template = "contextualItemsTypeError" if items else "contextualTypeError"
        kwargs["context"] = context
    elif items:
        template = "itemsTypeError"
    else:
        template = "typeError"

    return generateErrorMessage(template, **kwargs)


def validateType(
    value: Any,
    validTypes: type | tuple[type, ...] | UnionType,
    objectName: str,
    items=False,
) -> None:
    """Validate if a value matches any of the valid types.

    :param value: The value to be validated.
    :param validTypes: A single valid type or :class:`tuple` of valid
        types.
    :param objectName: The name of the object being validated.
    :param items: Whether to validate `value` items rather than `value`
        itself. Defaults to :obj:`False`.
    :raises TypeError: If `value` does not match any of the valid
        types.
    :raises ValueError: If  `items` is :obj:`True` and any `value` item
        does not match any of the valid types.

    Examples:

        >>> try:
        ...     error.validateType(123, str, "glyphName")
        ... except TypeError as e:
        ...     print(e)
        Expected 'glyphName' to be of type str, but got int

        >>> myList = ["uniE000", 1]
        >>> for item in myList:
        ...     try:
        ...         error.validateType(item, str, "myList", items=True)
        ...     except ValueError as e:
        ...         print(e)
        Items in 'myList' must be str, not int

    """
    if isinstance(validTypes, type):
        validTypes = (validTypes,)

    elif isinstance(validTypes, UnionType):
        validTypes = get_args(validTypes)

    if not isinstance(value, validTypes):
        if items:
            raise ValueError(
                generateTypeError(value, validTypes, objectName, items=items)
            )
        else:
            raise TypeError(generateTypeError(value, validTypes, objectName))


def suggestValue(
    value: str,
    possibilities: list[str] | tuple[str, ...],
    objectName: str,
    cutoff: float = 0.6,
    items=False,
) -> str:
    """Validate value and suggests a valid close match.

    If `items` is :obj:`True`, an alternate error message template is
    used to validate the values of items within an :term:`iterable`
    value rather than the value itself.

    :param value: The value to be validated.
    :param possibilities: A list of valid possibilities for the value.
    :param objectName: The name of the object being validated.
    :param cutoff: A float value representing the cutoff threshold for
        similarity matches. Defaults to ``0.6``.
    :param items: Whether to use items-specific template. Defaults
        to :obj:`False`.
    :raises ValueError: If the provided value is not found in
        `possibilities` or if `items` is :obj:`True` and `value` is
        not an iterable.

    Example:

        >>> try:
        ...     error.suggestValue(
        ...         "spiltStemUpSE", ["splitStemUpSE", "splitStemUpSW"],
        ...         "anchorName", cutoff=0.5)
        ... except ValueError as e:
        ...     print(e)
        Invalid value for 'anchorName': 'spiltStemUpSE'. Did you mean 'splitStemUpSE'?

    """
    if value in possibilities:
        return value
    closeMatches = difflib.get_close_matches(value, possibilities, 1, cutoff)
    if closeMatches:
        suggestion = closeMatches[0]
        template = "itemsValueError" if items else "valueError"
        raise ValueError(
            generateErrorMessage(
                template,
                "suggestion",
                objectName=objectName,
                value=value,
                suggestion=suggestion,
            )
        )

    raise ValueError(
        generateErrorMessage("valueError", objectName=objectName, value=value)
    )


def _listTypes(types: type | tuple[type, ...] | UnionType) -> str:
    # Represent a series of object type names as a string.

    if isinstance(types, type):
        types = (types,)

    elif isinstance(types, UnionType):
        types = get_args(types)

    if len(types) > 1:
        typeNames = (
            ", ".join(t.__name__ for t in types[:-1]) + f" or {types[-1].__name__}"
        )
    else:
        typeNames = types[0].__name__
    return typeNames



# pylint: disable=C0103, C0114
from __future__ import annotations
from typing import Any, TypeVar

# Generic
T = TypeVar("T")

PairType = tuple[T, T]
QuadrupleType = tuple[T, T, T, T]
QuintupleType = tuple[T, T, T, T, T]
SextupleType = tuple[T, T, T, T, T, T]
CollectionType = list[T] | tuple[T, ...]
PairCollectionType = list[T] | PairType[T]
QuadrupleCollectionType = list[T] | QuadrupleType[T]
SextupleCollectionType = list[T] | SextupleType[T]

JsonDict = dict[str, Any]
EngravingDefaultsInput = int | float | CollectionType[str]
EngravingDefaultsReturn = int | float | tuple[str, ...]



"""Utilities for normalizing various data types in SMufoLib.

This module extends :mod:`fontParts.base.normalizers` and provides
functions to normalize objects and data related to SMuFL fonts and
metadata requests. It includes normalization for:

    - :class:`.Font` objects
    - :class:`.Smufl` objects and attributes
    - :class:`.EngravingDefaults` objects and attributes
    - :class:`.Request` objects and attributes

These functions ensure that the data adheres to expected formats and
values, handling type validation and normalization as needed.

In addition to the documented functions below, it provides direct access
to the following external normalizers for convenience:

    - :func:`~fontParts.base.normalizers.normalizeBoolean`
    - :func:`~fontParts.base.normalizers.normalizeColor`
    - :func:`~fontParts.base.normalizers.normalizeVisualRounding`
    - :func:`~fontParts.base.normalizers.normalizeGlyph`

To import the module:

    >>> from smufolib import normalizers

"""

from __future__ import annotations
from typing import TYPE_CHECKING
from pathlib import Path

# ruff: noqa: F401
# pylint: disable=W0611
from fontParts.base.normalizers import normalizeInternalObjectType
from fontParts.base.normalizers import normalizeBoolean
from fontParts.base.normalizers import normalizeColor
from fontParts.base.normalizers import normalizeVisualRounding
from fontParts.base.normalizers import normalizeGlyph

from smufolib.utils import error
from smufolib.utils._annotations import (
    EngravingDefaultsInput,
    EngravingDefaultsReturn,
    CollectionType,
    PairCollectionType,
    PairType,
)

# ruff: noqa: F401
# pylint: enable=W0611
if TYPE_CHECKING:  # pragma: no cover
    from smufolib.objects.font import Font
    from smufolib.request import Request
    from smufolib.objects.smufl import Smufl
    from smufolib.objects.engravingDefaults import EngravingDefaults

# pylint: disable=C0103, C0415


# ----
# Font
# ----


def normalizeFont(value: Font) -> Font:
    """Normalize Font object.

    :param value: The value to normalize.
    :raises TypeError: If `value` is not the accepted type.

    """
    from smufolib.objects.font import Font

    return normalizeInternalObjectType(value, Font, "Font")


# -----
# Smufl
# -----


def normalizeClasses(value: CollectionType[str] | None) -> tuple[str, ...]:
    """Normalize smufl classes.

    :param value: The value to normalize.
    :raises TypeError: If `value` is not an accepted type.
    :raises ValueError:
        - If any `value` item does not normalize
          with :func:`normalizeSmuflName`.
        - If any `value` items are duplicates.

    """
    if value is None:
        return ()

    objectName = "Smufl.classes"
    error.validateType(value, (tuple, list), objectName)

    for val in value:
        error.validateType(val, str, objectName, items=True)
        for v in val.split("_"):
            normalizeSmuflName(v, items=True)

    duplicates = {v for v in value if value.count(v) > 1}
    if len(duplicates) != 0:
        raise ValueError(
            error.generateErrorMessage("duplicateItems", objectName=objectName)
        )

    return tuple(v for v in value if v)


def normalizeDescription(value: str | None) -> str | None:
    """Normalize smufl descriptions.

    :param value: The value to normalize.
    :raises TypeError: If `value` is not an accepted type.
    :raises ValueError: If `value` is an empty string.

    """
    if value is None:
        return None

    error.validateType(value, (str, type(None)), "Smufl.description")

    if len(value) == 0:
        raise ValueError(
            error.generateErrorMessage("emptyValue", objectName="Smufl.description")
        )

    return value


def normalizeDesignSize(value: int | None) -> int | None:
    """Normalize design size.

    :param value: The value to normalize.
    :raises TypeError: If `value` is not an accepted type.
    :raises ValueError: If `value` is less than ``10``.

    """
    error.validateType(value, (int, type(None)), "Smufl.designSize")

    if value is None:
        return None

    if value < 10:
        raise ValueError(
            error.generateErrorMessage(
                "valueTooLow", objectName="Smufl.designSize", value=10
            )
        )

    return value


def normalizeSizeRange(value: PairCollectionType[int] | None) -> PairType[int] | None:
    """Normalize design size.

    :param value: The value to normalize.
    :raises TypeError: If `value` or it's contents is not an accepted
        type.
    :raises ValueError:
        - If `value` does not contain a value pair.
        - If `value` is not an increasing range.
        - If `value` items do not normalize
          with :func:`normalizeDesignSize`, except not be :obj:`None`.

    """
    if value is None:
        return None

    objectName = "Smufl.sizeRange"

    error.validateType(value, (tuple, list, type(None)), objectName)

    if len(value) != 2:
        raise ValueError(
            error.generateErrorMessage("singleItem", objectName="Smufl.sizeRange")
        )

    start, end = value

    startNormalized = normalizeDesignSize(start)
    endNormalized = normalizeDesignSize(end)

    if start >= end:
        raise ValueError(
            error.generateErrorMessage("nonIncreasingRange", objectName=objectName)
        )

    return startNormalized, endNormalized  # type: ignore


def normalizeSmufl(value: Smufl) -> Smufl:
    """Normalize Smufl object.

    :param value: The value to normalize.
    :raises TypeError: If `value` is not the accepted type.

    """
    from smufolib.objects.smufl import Smufl

    return normalizeInternalObjectType(value, Smufl, "Smufl")


def normalizeSmuflName(value: str | None, items: bool = False) -> str | None:
    """Normalize smufl names.

    :param value: The value to normalize.
    :param items: Whether to normalize `value` items rather than `value`
        itself. Defaults to :obj:`False`.
    :raises TypeError: If `value` is not an accepted type.
    :raises ValueError:
        - If `value` is an empty string.
        - If `value` is value contains non-alphanumeric characters.
        - If `value` does not start with a lowercase letter or number.

    """
    if value is None:
        return None

    objectName = "Smufl.name"
    if items:
        objectName = "Smufl.classes"

    error.validateType(value, (str, type(None)), objectName)

    if not value:
        template = "emptyValueItems" if items else "emptyValue"
        raise ValueError(error.generateErrorMessage(template, objectName=objectName))

    for val in value:
        if not val.isalnum():
            template = "alphanumericValueItems" if items else "alphanumericValue"
            raise ValueError(
                error.generateErrorMessage(template, objectName=objectName)
            )

    if value[0].isupper():
        template = (
            "invalidInitialItemsCharacter" if items else "invalidInitialCharacter"
        )
        raise ValueError(error.generateErrorMessage(template, objectName=objectName))

    return value


# ------------------
# Engraving defaults
# ------------------


def normalizeEngravingDefaults(value: EngravingDefaults) -> EngravingDefaults:
    """Normalize EngravingDefaults object.

    :param value: The value to normalize.
    :raises TypeError: If `value` is not the accepted type.

    """
    from smufolib.objects.engravingDefaults import EngravingDefaults

    return normalizeInternalObjectType(value, EngravingDefaults, "EngravingDefaults")


def normalizeEngravingDefaultsAttr(
    name: str, value: EngravingDefaultsInput | None
) -> EngravingDefaultsReturn | None:
    """Normalize engraving defaults attribute.

    :param name: The attribute name to normalize.
    :param value: The attribute value to normalize.
    :raises TypeError: If any parameter value is not an accepted type.
    :raises AttributeError: If name is not a
        valid :class:`.EngravingDefaults` attribute name.
    :raises ValueError: If  ``"name=textFontFamily"`` and `value` is not
        a :class:`str` or empty.

    """
    from smufolib.objects.engravingDefaults import ENGRAVING_DEFAULTS_ATTRIBUTES

    if not isinstance(name, str):
        raise TypeError(
            error.generateTypeError(value=name, validTypes=str, objectName="name")
        )

    className = "EngravingDefaults"
    if name not in ENGRAVING_DEFAULTS_ATTRIBUTES:
        raise AttributeError(
            error.generateErrorMessage(
                "attributeError", objectName=className, attribute=name
            )
        )

    if value is None:
        return () if name == "textFontFamily" else None

    objectName = f"{className}.{name}"
    if name == "textFontFamily":
        if not isinstance(value, (tuple, list)):
            raise TypeError(
                error.generateTypeError(
                    value=value, validTypes=(tuple, list), objectName=objectName
                )
            )

        value = _normalizeStringTuple(objectName, value)

    elif not isinstance(value, (int, float)):
        raise TypeError(
            error.generateTypeError(
                value=value, validTypes=(int, float), objectName=objectName
            )
        )

    return value


def _normalizeStringTuple(
    objectName: str, value: CollectionType[str]
) -> tuple[str, ...]:
    # Normalize string tuple.
    error.validateType(value, (tuple, list), objectName)

    for val in value:
        error.validateType(val, str, objectName, items=True)
        if not val:
            raise ValueError(
                error.generateErrorMessage("emptyValueItems", objectName=objectName)
            )

    return tuple(value)


# -------
# Request
# -------


def normalizeRequest(value: Request) -> Request:
    """Normalize Request object.

    :param value: The value to normalize.
    :raises TypeError: If `value` is not the accepted type.

    """
    from smufolib.request import Request

    return normalizeInternalObjectType(value, Request, "Request")


def normalizeRequestPath(value: Path | str | None, parameter: str) -> str | None:
    """Normalize Request path.

    Relative paths are resolved automatically.

    :param value: The value to normalize.
    :param parameter: The name of the parameter being validated.
    :raises TypeError: If `value` is not the accepted type.

    """
    if value is None:
        return None
    error.validateType(value, (str, Path), parameter)
    if str(value).startswith("."):
        return str(Path(value).resolve())
    return str(value)



"""Miscellaneous general-purpose utility functions.

This module provides a set of utility functions for various common
tasks, including flattening nested iterables, summing tuples, and
extracting summary lines from docstrings. It also includes functions for
validating class attributes and string representations of floats,
printing verbose messages, as well as a a no-op function for
placeholder purposes.

To import the module:

    >>> from smufolib import stdUtils

"""

import operator
from typing import Any
from collections.abc import Generator, Iterable

# pylint: disable=C0103


def flatten(iterable: Iterable[Any], depth: int | None = None) -> Generator:
    """Flatten irregularly nested iterables of any depth.

    :param iterable: The :term:`iterable` to flatten.
    :param depth: The number of levels to flatten. `depth=None` employs
        maximum flattening. Defaults to :obj:`None`.

    Examples:

        >>> list(stdUtils.flatten([1, [2, [3, [4, 5]]]], depth=2))
        [1, 2, 3, [4, 5]]

        >>> list(stdUtils.flatten([1, [2, [3, [4, [5]]]]], depth=None))
        [1, 2, 3, 4, 5]

    """
    for item in iterable:
        if (
            not isinstance(item, Iterable)
            or isinstance(item, (str, bytes))
            or depth == 0
        ):
            yield item
        elif depth is None:
            yield from flatten(item)
        else:
            yield from flatten(item, depth - 1)


def addTuples(*tuples: tuple[int | float, ...]) -> tuple[int | float, ...]:
    r"""Sumize tuple values.

    :param \*tuples: tuples to add.

    Example:

        >>> stdUtils.addTuples((2, 4), (2, 4))
        (4, 8)

    """
    return tuple(map(sum, zip(*tuples)))


def getSummary(docstring: str | None) -> str | None:
    '''Extract the summary line from a docstring.

    :param docstring: The docstring from which to get the summary line.

    Example::

        >>> def showcaseGetSummary():
        ...     """Provide an example for the `getSummary()` function.
        ...
        ...     This function provides an example to show how to
        ...     extract a summary line from a docstring.
        ...
        ...     """
        ...
        >>> docstring = showcaseGetSummary.__doc__
        >>> stdUtils.getSummary(docstring)
        "Give example for `getSummary()` function."

    '''
    if docstring is None:
        return None

    return docstring.split("\n", maxsplit=1)[0]


def isFloat(string: str) -> bool:
    """Check if string represents a :class:`float`.

    :param string: The string to check.

    Examples:

        >>> stdUtils.isFloat("3.14")
        True

        >>> stdUtils.isFloat("314")
        False

    """
    if "." not in string:
        return False
    try:
        float(string)
        return True
    except ValueError:
        return False


def validateAttr(obj, attributes: Iterable[str] | None = None) -> bool:
    """Validate `obj` based on whether the given attributes have been set.

    This function returns :obj:`False` if any given attribute is not :obj:`None`,
    otherwise itreturns :obj:`True`. It supports validation of nested
    attributes, e.g.: ``"attr1.attr2.attr3"``.

    :param obj: The object to validate.
    :param attributes: The attribute names to check.

    Example:

        >>> class MyClass:
        ...     def __init__(self):
        ...         self.attr1 = 1
        ...         self.attr2 = None
        ...         self.attr3 = False
        ...
        >>> obj = MyClass()
        >>> stdUtils.validateAttr(obj, "attr1")
        True
        >>> stdUtils.validateAttr(obj, "attr2")
        False
        >>> stdUtils.validateAttr(obj, "attr3")
        True
        >>> stdUtils.validateAttr(obj, ["attr1", "attr2"])
        False

    """
    if not attributes:
        return False

    if isinstance(attributes, str):
        attributes = (attributes,)

    return all(operator.attrgetter(attr)(obj) is not None for attr in attributes)


def doNothing(*args: Any, **kwargs: Any) -> None:
    # pylint: disable=W0613
    r"""Do nothing at all.

    This function accepts any arguments but does nothing with them.

    :param \*args: Positional arguments.
    :param \**kwargs: Keyword arguments.

    Example:

        >>> stdUtils.doNothing(1, 2, 3)
        >>> stdUtils.doNothing(a=1, b=2)

    """


def verbosePrint(message: str, verbose: bool, *args: Any, **kwargs: Any) -> None:
    r"""Print a message if `verbose` is :obj:`True`.

    This function behaves like the built-in :func:`print` function, but
    only prints the message if `verbose` is :obj:`True`. If `verbose`
    is :obj:`False`, it does nothing.

    :param message: The message to be printed.
    :param verbose: A flag indicating whether to print the message.
    :param \*args: Additional positional arguments passed to
        the :func:`print` function.
    :param \**kwargs: Additional keyword arguments passed to
        the :func:`print` function.

    Example:

        >>> stdUtils.verbosePrint("Hello, world!", True)
        Hello, world!
        >>> stdUtils.verbosePrint("Hello, world!", False)

    """
    if verbose:
        print(message, *args, **kwargs)
    else:
        doNothing(message, *args, **kwargs)



# pylint: disable=C0114
from __future__ import annotations

from fontParts.fontshell.glyph import RGlyph
from smufolib.objects.smufl import Smufl


class Glyph(RGlyph):
    """SMufoLib environment implementation of :class:`fontParts.base.BaseGlyph`.

    .. versionchanged:: 0.6.0

       The ``__repr__`` now includes the canonoical SMuFL glyph name (if present) in
       square brackets.

    This object is typically accessed through a :class:`.Font` object's inherent glyph
    dictionary. To instantiate the SMuFL glyph named U+E050 (*gClef*):

        >>> glyph = font["uniE050"]

    A glyph may also be instantiated independently and assigned to a font or glyph
    later:

        >>> glyph = Glyph()  # doctest: +SKIP

    """

    def _reprContents(self) -> list[str]:
        # Adds bracketed smufl name if available
        contents = [f"'{self.name}'"]
        smuflName = getattr(self.smufl, "name", None)
        if smuflName:
            contents.append(f"['{smuflName}']")
        if self.layer is not None:
            contents.append(f"('{self.layer.name}')")
        return contents

    def _set_name(self, value: str) -> None:
        # Set the name of the glyph and update :attr:`.Font.lib` if necessary.
        if self.font is not None:
            namesDict = self.smufl.names
            smuflName = self.smufl.name
            if namesDict and smuflName and self.naked().name in namesDict.values():
                namesDict[smuflName] = value
        self.naked().name = value

    @property
    def smufl(self) -> Smufl:
        """Glyph-specific instance of :class:`~smufolib.objects.smufl.Smufl`."""
        return Smufl(glyph=self)



#!/usr/bin/env python3
"""Check SMuFL glyph anchors against reference font metadata.

This script checks font anchors against reference font metadata, identifying missing or
superfluous glyph anchors according to the SMuFL standard.

Discrepancies are printed to console, and glyphs may be marked with a specified color
value. Glyphs without annotated smufl names or glyph names will be skipped.

The script requires SMufoLib to be installed in its execution environment. It can be
used from the command line or as a Python module. See the :ref:`check-anchors-cli` and
:ref:`check-anchors-python` sections below for usage details.

"""

from __future__ import annotations
from typing import Any
import argparse
from pathlib import Path

from tqdm import tqdm

from smufolib import (
    Font,
    Request,
    cli,
    config,
    stdUtils,
)
from smufolib.utils.scriptUtils import normalizeColor as _normalizeColor
from smufolib.utils.scriptUtils import normalizeFont as _normalizeFont
from smufolib.utils.scriptUtils import normalizeJsonDict as _normalizeJsonDict
from smufolib.utils.scriptUtils import normalizeRequest as _normalizeRequest

# Type aliases
JsonDict = dict[str, Any]
ColorValue = int | float
ColorTuple = tuple[ColorValue, ColorValue, ColorValue, ColorValue]

# Configuration
CONFIG = config.load()

# Parameter defaults
FONT_DATA = Request(
    CONFIG["metadata.paths"]["font"],
    CONFIG["metadata.fallbacks"]["font"],
)
MARK = False
MARK_COLOR = CONFIG["color.marks"]["mark1"]
VERBOSE = False

# pylint: disable=invalid-name


def checkAnchors(
    font: Font | Path | str,
    fontData: Request | Path | str = FONT_DATA,
    mark: bool = MARK,
    color: ColorTuple | None = MARK_COLOR,
    verbose: bool = VERBOSE,
) -> None:
    """Check font anchors against SMuFL reference data (Python API).

    :param font: Target :class:`.Font` object or path to font file.
    :param fontData: Request for or path to reference font metadata file. Defaults to
        :class:`smufolib.request.Request` passing :confval:`metadata.paths.font` and
        :confval:`metadata.fallbacks.font`.
    :param mark: Mark discrepant glyphs. Defaults to :obj:`False`.
    :param color: Color to use when marking glyphs. Ignored unless ``mark=True``.
        Defaults to :confval:`color.marks.mark1`.
    :param verbose: Make output verbose. Defaults to :obj:`False`.
    :raises TypeError: If any parameter value is not the expected type or if `mark` is
        :obj:`True` while `color` is :obj:`None`.

    """
    print("Starting...")

    names = {}
    fontAnchors = {}
    referenceAnchors = {}
    font = _normalizeFont(font)
    metadata = _normalizeJsonDict(_normalizeRequest(fontData).json())
    color = _normalizeColor(color, mark)

    # Build dicts of glyph names and anchors indexed on smufl names.
    stdUtils.verbosePrint("\nCompiling font anchors...", verbose)
    for glyph in font if verbose else tqdm(font):
        names[glyph.smufl.name] = glyph.name
        if glyph.anchors:
            fontAnchors[glyph.smufl.name] = [a.name for a in glyph.anchors]

    if not fontAnchors:
        stdUtils.verbosePrint("\nNo font anchors found.", verbose)
    else:
        # Build dict of reference anchors indexed on smufl names.
        stdUtils.verbosePrint("\nCompiling reference anchors...", verbose)
        for name, anchors in metadata["glyphsWithAnchors"].items():
            if name in fontAnchors:
                referenceAnchors[name] = anchors.keys()

        # Compare reference anchors to font anchors to get deficit.
        stdUtils.verbosePrint("\nGlyphs with missing anchors:", verbose)
        result = _evaluate(referenceAnchors, fontAnchors, names, verbose)
        if mark and color and result:
            for name in result:
                glyph = font[names[name]]
                glyph.markColor = color

        # Compare font anchors to reference anchors to get surplus.
        stdUtils.verbosePrint("\nGlyphs with superfluous anchors:", verbose)
        result = _evaluate(fontAnchors, referenceAnchors, names, verbose)
        if mark and color and result:
            for name in result:
                glyph = font[names[name]]
                glyph.markColor = color

        print("\nDone.")


def main() -> None:
    """Command line entry point."""
    args = _parseArgs()
    checkAnchors(
        args.font,
        fontData=args.fontData,
        mark=args.mark,
        color=args.color,
        verbose=args.verbose,
    )


def _evaluate(
    test: JsonDict,
    reference: JsonDict,
    names: dict[str, str],
    verbose: bool,
) -> list[str]:
    # Compare sources and print results.
    findings: list[str] = []

    for name, anchors in test.items():
        if name not in reference:
            continue

        for anchor in anchors:
            if anchor in reference[name]:
                continue

            if name not in findings:
                findings.append(name)
                stdUtils.verbosePrint(f"\t'{names[name]}' ('{name}'):", verbose)
            stdUtils.verbosePrint(f"\t\t'{anchor}'", verbose)

    if not findings:
        stdUtils.verbosePrint("\tNone", verbose)

    return findings


def _parseArgs() -> argparse.Namespace:
    # Parse command line arguments and options.
    parser = cli.commonParser(
        "font",
        description=stdUtils.getSummary(__doc__),
        fontData=FONT_DATA,
        mark=MARK,
        color=MARK_COLOR,
        verbose=VERBOSE,
    )
    return parser.parse_args()


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
"""Remove SMuFL font data.

.. warning::

    This script will permanently delete data. Always back up your font file before use.

This script deletes SMuFL-specific metadata and glyph anchors from a UFO font file.

The following attribute and anchor names may be specified for inclusion or exclusion:

    Attributes
        - `designSize`
        - `sizeRange`
        - `engravingDefaults`
        - `name`
        - `classes`
        - `description`
        - `spaces`

    Anchors
        - `splitStemUpSE`
        - `splitStemUpSW`
        - `splitStemDownNE`
        - `splitStemDownNW`
        - `stemUpSE`
        - `stemDownNW`
        - `stemUpNW`
        - `stemDownSW`
        - `nominalWidth`
        - `numeralTop`
        - `numeralBottom`
        - `cutOutNE`
        - `cutOutSE`
        - `cutOutSW`
        - `cutOutNW`
        - `graceNoteSlashSW`
        - `graceNoteSlashNE`
        - `graceNoteSlashNW`
        - `graceNoteSlashSE`
        - `repeatOffset`
        - `noteheadOrigin`
        - `opticalCenter`

The script requires SMufoLib to be installed in its execution environment. It can be
used from the command line or as a Python module. See the :ref:`clean-font-cli` and
:ref:`clean-font-python` sections below for usage details.

"""

from collections.abc import Iterable
import argparse
from pathlib import Path

from tqdm import tqdm

from smufolib import Font
from smufolib.objects.smufl import ANCHOR_NAMES, FONT_ATTRIBUTES, GLYPH_ATTRIBUTES
from smufolib import cli
from smufolib.utils import error, stdUtils
from smufolib.utils.scriptUtils import normalizeFont as _normalizeFont


# Parameter defaults
EXCLUDE = None
VERBOSE = False

# pylint: disable=C0103


def cleanFont(
    font: Font | Path | str,
    include: Iterable,
    exclude: Iterable | None = EXCLUDE,
    verbose: bool = VERBOSE,
):
    """Delete Smufl-specific attribute values (Python API).

    Use the `include` and `exclude` keyword arguments to control which attributes and
    anchors are removed.

    To avoid unintentional data loss, `include` must be specified. Setting `include="*"`
    will delete all SMuFL metadata and anchors.

    Refer to the list of valid keys in the main docstring.

    :param font: Object or path to target :class:`~smufolib.objects.font.Font`.
    :param include: Items to be deleted. May be ``"*"`` (all), an individual attribute
        or anchor name as a :class:`str` or :class:`tuple` of several.
    :param exclude: Items to be preserved if ``ìnclude="*"``. Defaults to :obj:`None`.
    :param verbose: Make output verbose. Defaults to :obj:`False`.
    :raises TypeError: If any parameter value is not the expected type.
    :raises ValueError: If any item in `include` or `exclude` is invalid.

    """
    print("Starting...")

    font = _normalizeFont(font)
    itemsToClean = _buildItemsDict(include, exclude)

    # Clean font attributes
    stdUtils.verbosePrint("\nCleaning attributes for font:", verbose)
    for attr in itemsToClean["fontAttributes"]:
        if font.smufl.engravingDefaults and attr == "engravingDefaults":
            font.smufl.engravingDefaults.clear()
        elif attr == "spaces":
            setattr(font.smufl, attr, False)
        else:
            setattr(font.smufl, attr, None)
        stdUtils.verbosePrint(f"\t'{attr}'", verbose)

    for glyph in font if verbose else tqdm(font):
        # Clean glyph attributes
        glyphAttributesCleaned = False
        for attr in itemsToClean["glyphAttributes"]:
            if not getattr(glyph.smufl, attr):
                continue

            if not glyphAttributesCleaned:
                stdUtils.verbosePrint(
                    f"\nCleaning attributes from glyph '{glyph.name}':", verbose
                )
                glyphAttributesCleaned = True

            setattr(glyph.smufl, attr, None)
            stdUtils.verbosePrint(f"\t'{attr}'", verbose)

        # Clean Anchors
        anchorsCleaned = False
        for anchor in glyph.anchors:
            if anchor.name not in itemsToClean["anchors"]:
                continue

            if not anchorsCleaned:
                stdUtils.verbosePrint(
                    f"\nCleaning anchors from glyph '{glyph.name}':", verbose
                )
                anchorsCleaned = True

            glyph.removeAnchor(anchor.index)
            stdUtils.verbosePrint(f"\t'{anchor.name}'", verbose)

    stdUtils.verbosePrint("\nSaving font...", verbose)
    font.save()

    print("\nDone.")


def main() -> None:
    """Command line entry point."""
    args = _parseArgs()
    cleanFont(args.font, args.include, exclude=args.exclude, verbose=args.verbose)


def _buildItemsDict(include, exclude):
    # Build dict of attribute and anchor items to remove.

    itemsToClean = {"fontAttributes": [], "glyphAttributes": [], "anchors": []}

    allItems = FONT_ATTRIBUTES | GLYPH_ATTRIBUTES | ANCHOR_NAMES
    exclude = () if exclude is None else exclude
    exclude = (exclude,) if isinstance(exclude, str) else exclude

    for item in exclude:
        error.suggestValue(item, allItems, "exclude")

    if include in ("*", ["*"]):
        include = allItems
    elif isinstance(include, str):
        include = (include,)

    for item in include:
        if item in exclude:
            continue

        if item in FONT_ATTRIBUTES:
            itemsToClean["fontAttributes"].append(item)
        elif item in GLYPH_ATTRIBUTES:
            itemsToClean["glyphAttributes"].append(item)
        elif item in ANCHOR_NAMES:
            itemsToClean["anchors"].append(item)
        else:
            error.suggestValue(item, allItems, "include")

    return itemsToClean


def _parseArgs() -> argparse.Namespace:
    # Parse command line arguments and options.
    parser = cli.commonParser(
        "font",
        "include",
        description=stdUtils.getSummary(cleanFont.__doc__),
        exclude=EXCLUDE,
        verbose=VERBOSE,
    )
    return parser.parse_args()


if __name__ == "__main__":
    main()



"""Measurement utilities for calculating glyph geometry.

This module provides functions to retrieve, measure and inspect various parts of glyphs.
It also defines:

- :data:`ENGRAVING_DEFAULTS_MAPPING`, the default association between ruler functions
  and glyphs used for calculating :class:`.EngravingDefaults` attribute values.
- :data:`DISPATCHER`, for dynamically executing measurement operations by function name.

To import the module:

    >>> from smufolib import rulers

"""

from __future__ import annotations
from typing import TYPE_CHECKING, Generator
from collections.abc import Callable
from math import dist

from smufolib.utils import converters

if TYPE_CHECKING:  # pragma: no cover
    from fontParts.fontshell.point import RPoint
    from fontParts.fontshell.segment import RSegment
    from fontParts.fontshell.contour import RContour
    from smufolib.objects.glyph import Glyph

Bounds = tuple[int | float, int | float, int | float, int | float]
Remapping = dict[str, dict[str, str]]
RulerType = Callable[["Glyph"], int | float | None]

TOLERANCE: int | float = 6
TYPES: str | tuple[str, ...] = ("line", "curve", "qcurve")

#: Default mapping of rulers and glyphs to :class:`EngravingDefaults` attributes.
ENGRAVING_DEFAULTS_MAPPING: Remapping = {
    "arrowShaftThickness": {
        "ruler": "xStrokeWidthAtOrigin",
        "glyph": "uniEB60",
    },
    "barlineSeparation": {
        "ruler": "xDistanceBetweenContours",
        "glyph": "uniE031",
    },
    "beamSpacing": {
        "ruler": "yDistanceBetweenContours",
        "glyph": "uniE1F9",
    },
    "beamThickness": {
        "ruler": "glyphBoundsHeight",
        "glyph": "uniE1F7",
    },
    "bracketThickness": {
        "ruler": "xStrokeWidthAtOrigin",
        "glyph": "uniE003",
    },
    "dashedBarlineDashLength": {
        "ruler": "yStrokeWidthAtMinimum",
        "glyph": "uniE036",
    },
    "dashedBarlineGapLength": {
        "ruler": "yDistanceBetweenContours",
        "glyph": "uniE036",
    },
    "dashedBarlineThickness": {
        "ruler": "glyphBoundsWidth",
        "glyph": "uniE036",
    },
    "hairpinThickness": {
        "ruler": "wedgeArmStrokeWidth",
        "glyph": "uniE53E",
    },
    "hBarThickness": {
        "ruler": "yStrokeWidthAtMinimum",
        "glyph": "uniE4F0",
    },
    "legerLineExtension": {
        "ruler": "glyphBoundsXMinAbs",
        "glyph": "uniE022",
    },
    "legerLineThickness": {
        "ruler": "glyphBoundsHeight",
        "glyph": "uniE022",
    },
    "lyricLineThickness": {
        "ruler": "glyphBoundsHeight",
        "glyph": "uniE010",
    },
    "octaveLineThickness": {
        "ruler": "glyphBoundsHeight",
        "glyph": "uniE010",
    },
    "pedalLineThickness": {
        "ruler": "glyphBoundsHeight",
        "glyph": "uniE010",
    },
    "repeatBarlineDotSeparation": {
        "ruler": "xDistanceStemToDot",
        "glyph": "uniE040",
    },
    "repeatEndingLineThickness": {
        "ruler": "xStrokeWidthAtOrigin",
        "glyph": "uniE030",
    },
    "slurEndpointThickness": {
        "ruler": "xStrokeWidthAtOrigin",
        "glyph": "uniE1FD",
    },
    "slurMidpointThickness": {
        "ruler": "yStrokeWidthAtMinimum",
        "glyph": "uniE1FD",
    },
    "staffLineThickness": {
        "ruler": "glyphBoundsHeight",
        "glyph": "uniE010",
    },
    "stemThickness": {
        "ruler": "glyphBoundsWidth",
        "glyph": "uniE210",
    },
    "subBracketThickness": {
        "ruler": "xStrokeWidthAtOrigin",
        "glyph": "uniE030",
    },
    "textEnclosureThickness": {
        "ruler": "glyphBoundsHeight",
        "glyph": "uniE010",
    },
    "textFontFamily": {},
    "thickBarlineThickness": {
        "ruler": "glyphBoundsWidth",
        "glyph": "uniE034",
    },
    "thinBarlineThickness": {
        "ruler": "glyphBoundsWidth",
        "glyph": "uniE030",
    },
    "thinThickBarlineSeparation": {
        "ruler": "xDistanceBetweenContours",
        "glyph": "uniE032",
    },
    "tieEndpointThickness": {
        "ruler": "xStrokeWidthAtOrigin",
        "glyph": "uniE1FD",
    },
    "tieMidpointThickness": {
        "ruler": "yStrokeWidthAtMinimum",
        "glyph": "uniE1FD",
    },
    "tupletBracketThickness": {
        "ruler": "xStrokeWidthAtOrigin",
        "glyph": "uniE1FE",
    },
}

# ------
# Rulers
# ------


def glyphBoundsHeight(glyph: Glyph) -> int | float | None:
    """Return height of the glyph's bounding box.

    :param glyph: Source :class:`.Glyph` of contours to measure.

    Example:

        >>> glyph = font["uniE050"]
        >>> rulers.glyphBoundsHeight(glyph)
        1801

    """
    boundsList = [c.bounds for c in getGlyphContours(glyph)]
    bounds = combineBounds(boundsList)
    if not bounds:
        return None
    return converters.toIntIfWhole(bounds[3] - bounds[1])


def glyphBoundsWidth(glyph: Glyph) -> int | float | None:
    """Return width of the glyph's bounding box.

    :param glyph: Source :class:`.Glyph` of contours to measure.

    Example:

        >>> glyph = font["uniE050"]
        >>> rulers.glyphBoundsWidth(glyph)
        648

    """
    boundsList = [c.bounds for c in getGlyphContours(glyph)]
    bounds = combineBounds(boundsList)
    if not bounds:
        return None
    return converters.toIntIfWhole(bounds[2] - bounds[0])


def glyphBoundsXMinAbs(glyph: Glyph) -> int | float | None:
    """Return absolute value of glyph's *xMin* bound.

    :param glyph: Source :class:`.Glyph` of contours to measure.

    Example::

        >>> glyph = font["uniE022"]
        >>> rulers.glyphBoundsXMinAbs(glyph)
        80

    """
    boundsList = [c.bounds for c in getGlyphContours(glyph)]
    bounds = combineBounds(boundsList)
    if not bounds:
        return None
    return converters.toIntIfWhole(abs(bounds[0]))


def xDistanceStemToDot(glyph: Glyph) -> int | float | None:
    """Measure horizontal distance between a stem and a dot contour.

    The contours may be placed on either side of each other.

    :param glyph: Source :class:`.Glyph` of contours to measure.

    Example::

        >>> glyph = font["uniE040"]
        >>> rulers.glyphBoundsXMinAbs(glyph)
        63

    """
    dotPoints = [
        p for p in getGlyphPoints(glyph) if all(s.type != "line" for s in p.contour)
    ]
    stemPoints = list(getGlyphPoints(glyph, "line"))

    if not stemPoints or not dotPoints:
        return None

    # reference point should be the point in the dot contour facing the stem.
    referencePoint = min(dotPoints, key=lambda p: p.x)
    closestStemPoint = max(stemPoints, key=lambda p: p.x)
    if referencePoint.x < closestStemPoint.x:
        referencePoint = max(dotPoints, key=lambda p: p.x)
        closestStemPoint = min(stemPoints, key=lambda p: p.x)

    return converters.toIntIfWhole(abs(closestStemPoint.x - referencePoint.x))


def xDistanceBetweenContours(glyph: Glyph) -> int | float | None:
    """Measure horizontal distance between two adjacent contours in a glyph.

    :param glyph: Source :class:`.Glyph` of contours to measure.

    Example::

        >>> glyph = font["uniE031"]
        >>> rulers.xDistanceBetweenContours(glyph)
        85

    """
    points = sorted(getGlyphPoints(glyph), key=lambda p: p.x)
    if not points:
        return None

    # Reference point should be the right point of the left contour.
    referencePoint = max(points[0].contour.points, key=lambda p: p.x)

    for point in points:
        if point.contour.index == referencePoint.contour.index:
            continue

        return converters.toIntIfWhole(abs(point.x - referencePoint.x))

    return None


def yDistanceBetweenContours(glyph: Glyph) -> int | float | None:
    """Measure vertical distance between two adjacent contours in a glyph.

    :param glyph: Source :class:`.Glyph` of contours to measure.

    Example::

        >>> glyph = font["uniE036"]
        >>> rulers.yDistanceBetweenContours(glyph)
        119

    """
    points = sorted(getGlyphPoints(glyph), key=lambda p: p.y)
    if not points:
        return None

    # Reference point should be the top point of the bottom contour.
    referencePoint = max(points[0].contour.points, key=lambda p: p.y)

    for point in points:
        if point.contour.index == referencePoint.contour.index:
            continue

        return converters.toIntIfWhole(abs(point.y - referencePoint.y))

    return None


def xStrokeWidthAtOrigin(glyph: Glyph) -> int | float | None:
    """Measure horizontal distance between aligned points closest to origin.

    :param glyph: Source :class:`.Glyph` of contours to measure.

    Example::

        >>> glyph = font["uniE1FE"]
        >>> rulers.xStrokeWidthAtOrigin(glyph)
        32

    """
    points = sorted(getGlyphPoints(glyph), key=lambda p: abs(p.x) + abs(p.y))
    if not points:
        return None

    referencePoint, *rest = points

    for point in rest:
        if point.contour.index != referencePoint.contour.index:
            continue
        if point.x == referencePoint.x:
            continue
        if not areAlligned((referencePoint, point), axis="y"):
            continue
        return converters.toIntIfWhole(abs(point.x - referencePoint.x))

    return None


def yStrokeWidthAtMinimum(glyph: Glyph) -> int | float | None:
    """Measure vertical distance between aligned low-points in the glyph.

    :param glyph: Source :class:`.Glyph` of contours to measure.

    Example::

        >>> glyph = font["uniE1FD"]
        >>> rulers.yStrokeWidthAtMinimum(glyph)
        50

    """

    points = sorted(getGlyphPoints(glyph), key=lambda p: p.y)
    if not points:
        return None

    referencePoint, *rest = points

    for point in rest:
        if point.contour.index != referencePoint.contour.index:
            continue
        if point.y == referencePoint.y:
            continue
        if not areAlligned((referencePoint, point), axis="x"):
            continue
        return converters.toIntIfWhole(abs(point.y - referencePoint.y))

    return None


def wedgeArmStrokeWidth(glyph: Glyph):
    """Measure thickness of arm in a wedge-shaped glyph.

    :param glyph: Source :class:`.Glyph` of contours to measure.

    Example::

        >>> glyph = font["uniE1FD"]
        >>> rulers.yStrokeWidthAtMinimum(glyph)
        50

    """

    def euclidean(p1, p2):
        return dist((p1.x, p1.y), (p2.x, p2.y))

    points = sorted(getGlyphPoints(glyph, "line"), key=lambda p: p.y)

    referencePoint = max(points, key=lambda p: p.x)

    candidates = [p for p in points if p is not referencePoint]
    nearest = min(candidates, key=lambda p: euclidean(p, referencePoint))

    return round(euclidean(referencePoint, nearest))


def areAlligned(
    points: tuple[RPoint, ...], axis: str, tolerance: int | float = TOLERANCE
) -> bool:
    """Return :obj:`True` if the specified `points` are aligned on `axis`.

    :param points: The points to check for alignment.
    :param axis: The axis (``"x"`` or ``"y"``) accross which to check for alignment.
    :param tolerance: The tolerance for misalignment to apply in font units.
        Defaults to ``6``.

    Example:

        >>> glyph = font["uniE050"]
        >>> point1, point2 = glyph[0].points[:2]
        >>> rulers.areAlligned((point1, point2), axis="x", tolerance=3)
        False

    """

    def withinRange(value: int | float, reference: int | float) -> bool:
        # Check if values are within range of each other +/- the tolerance level.
        return value - tolerance <= reference <= value + tolerance

    first, *rest = points
    reference = getattr(first, axis)
    return all(withinRange(getattr(v, axis), reference) for v in rest)


def getGlyphContours(
    glyph: Glyph, includeComponents: bool = True
) -> Generator[RContour]:
    """Return all contours in the glyph, including component references.

    :param glyph: The :class:`.Glyph` containing the contours to retrieve.
    :param includeComponents: Whether to include any referenced contours in the glyph.
        Defaults to :obj:`True`.

    Example:

        >>> glyph = font["uniE050"]
        >>> rulers.getGlyphContours(glyph)  # doctest: +ELLIPSIS
        <generator object getGlyphContours.<locals>.<genexpr> at 0x...>

    """
    if glyph.components and includeComponents:
        tempGlyph = glyph.copy()
        font = glyph.font
        tempName = "com.smufolib.temp"
        font.insertGlyph(tempGlyph, name=tempName)
        tempAssigned = font[tempName]
        tempAssigned.decompose()
        contours = (c for c in tempAssigned)
        font.removeGlyph(tempName)
        return contours
    return (c for c in glyph)


def getGlyphSegments(
    glyph, types: str | tuple[str, ...] = TYPES, includeComponents: bool = True
) -> Generator[RSegment]:
    """Return all segments in the `glyph` matching given `types`.

    Any segments referenced by components may be included.

    :param glyph: The :class:`.Glyph` containing the segments to retrieve.
    :param types: The :attr:`~fontParts.base.BaseSegment.type` values to include.
        Defaults to ``("line", "curve", "qcurve")``.
    :param includeComponents: Whether to include any referenced contours in the glyph.

    Example:

        >>> glyph = font["uniE050"]
        >>> rulers.getGlyphSegments(glyph)  # doctest: +ELLIPSIS
        <generator object getGlyphSegments.<locals>.<genexpr> at 0x...>

    """
    contours = getGlyphContours(glyph, includeComponents=includeComponents)
    return (s for sts in contours for s in sts if s.type in types or s.type == types)


def getGlyphPoints(
    glyph, types: str | tuple[str, ...] = TYPES, includeComponents: bool = True
) -> Generator[RPoint]:
    """Return all points in the `glyph` matching given `types`.

    Any points referenced by components may be included.

    :param glyph: The :class:`.Glyph` containing the points to retrieve.
    :param types: The :attr:`~fontParts.base.BasePoint.type` values to include.
        Defaults to ``("line", "curve", "qcurve")``.
    :param includeComponents: Whether to include any referenced contours in the glyph.

    Example:

        >>> glyph = font["uniE050"]
        >>> rulers.getGlyphPoints(glyph)  # doctest: +ELLIPSIS
        <generator object getGlyphPoints.<locals>.<genexpr> at 0x...>

    """

    segments = getGlyphSegments(glyph, includeComponents=includeComponents, types=types)
    return (p for pts in segments for p in pts if p.type in types or p.type == types)


def hasHorizontalOffCurve(point: RPoint) -> bool:
    """Return :obj:`True` if `point` has a predominantly horizontal off-curve.

    An off-curve control point is considered horizontal if its *x-difference* from the
    on-curve point is greater than its *y-difference*.

    :param point: The on-curve point to check.

    Example:

        >>> glyph = font["uniE260"]
        >>> point = glyph[0].points[0]
        >>> rulers.hasHorizontalOffCurve(point)
        True

    """
    segment = getParentSegment(point)
    if segment and point.type in ("curve", "qcurve"):
        for offCurve in segment.offCurve:
            onX, onY = point.position
            offX, offY = offCurve.position
            xDiff = abs(onX - offX)
            yDiff = abs(onY - offY)
            if xDiff > yDiff:
                return True
    return False


def hasVerticalOffCurve(point: RPoint) -> bool:
    """Return :obj:`True` if `point` has a predominantly vertical off-curve.

    An off-curve control point is considered vertical if its *y-difference* from the
    on-curve point is greater than its *x-difference*.

    :param point: The on-curve point to check.

    Example:

        >>> glyph = font["uniE260"]
        >>> point = glyph[0].points[0]
        >>> rulers.hasVerticalOffCurve(point)
        False

    """
    segment = getParentSegment(point)
    if segment and point.type in ("curve", "qcurve"):
        for offCurve in segment.offCurve:
            onX, onY = point.position
            offX, offY = offCurve.position
            xDiff = abs(onX - offX)
            yDiff = abs(onY - offY)
            if xDiff < yDiff:
                return True
    return False


def getParentSegment(point: RPoint) -> RSegment | None:
    """Get the segment which the `point` belongs to.

    :param point: The point to find the parent segment for.

    Example:

        >>> glyph = font["uniE050"]
        >>> point = glyph[0].points[0]
        >>> rulers.getParentSegment(point)  # doctest: +ELLIPSIS
        <RSegment line index='3' at ...>

    """
    match = next(
        (s for s in point.contour.segments if point.contour and point in s), None
    )
    return match


def combineBounds(boundsList: list[Bounds]) -> Bounds | None:
    """Combine a list of bounds into one bounding box.

    Example:

        >>> glyph1, glyph2 = font["uniE050"], font["uniE260"]
        >>> boundsList = [glyph1.bounds, glyph2.bounds]
        >>> rulers.combineBounds(boundsList)
        (-50, -634, 648, 1167)

    """
    if not boundsList:
        return None

    xMins, yMins, xMaxs, yMaxs = zip(*boundsList)
    return (min(xMins), min(yMins), max(xMaxs), max(yMaxs))


# ----------
# Dispatcher
# ----------

#: Dispatch the different ruler functions by their :class:`str` name.
DISPATCHER: dict[str, RulerType] = {
    "glyphBoundsHeight": glyphBoundsHeight,
    "glyphBoundsWidth": glyphBoundsWidth,
    "glyphBoundsXMinAbs": glyphBoundsXMinAbs,
    "xDistanceStemToDot": xDistanceStemToDot,
    "xDistanceBetweenContours": xDistanceBetweenContours,
    "xStrokeWidthAtOrigin": xStrokeWidthAtOrigin,
    "yDistanceBetweenContours": yDistanceBetweenContours,
    "yStrokeWidthAtMinimum": yStrokeWidthAtMinimum,
    "wedgeArmStrokeWidth": wedgeArmStrokeWidth,
}



#!/usr/bin/env python3
"""Calculate engraving defaults from glyph contours.

This script calculates and sets attribute values for the :class:`.EngravingDefaults`
class based on glyph dimensions, registration, metrics, and point placement.

It offers options to override automatic contour measurements and reassign attributes to
different ruler functions or glyphs. Refer to :data:`.ENGRAVING_DEFAULTS_MAPPING` for a
comprehensive list of attributes and their default corresponding glyphs and assigned
ruler functions.

.. note::

    The setting for :attr:`.textFontFamily` must be set manually within the `override`
    parameter.

The script requires SMufoLib to be installed in its execution environment. It can be
used from the command line or as a Python module. See the
:ref:`calculate-engraving-defaults-cli` and :ref:`calculate-engraving-defaults-python`
sections below for usage details.



.. versionchanged:: 0.5.0

    The `referenceIndex` parameter has been removed from all ruler functions and is no
    longer needed when providing remappings.

    The ruler functions have been removed from this script and are now available in the
    :mod:`~smufolib.utils.rulers` module.

"""

from __future__ import annotations
from collections.abc import Callable
import argparse
import json
import textwrap
from pathlib import Path

from tqdm import tqdm

from smufolib import (
    Font,
    Glyph,
    cli,
    config,
    error,
    normalizers,
    stdUtils,
)
from smufolib.utils.rulers import ENGRAVING_DEFAULTS_MAPPING, DISPATCHER
from smufolib.utils.scriptUtils import normalizeFont as _normalizeFont


Exclude = tuple[str] | list[str]
OverrideValue = int | float | tuple[str, ...] | None
Override = dict[str, OverrideValue]
Remapping = dict[str, dict[str, str]]
RulerType = Callable[[Glyph], int | float | None]

CONFIG = config.load()

# Parameter defaults
EXCLUDE = None
OVERRIDE = None
SPACES = False
REMAP = None
VERBOSE = False


# pylint: disable=R0913, C0103, R0914


def calculateEngravingDefaults(
    font: Font | Path | str,
    exclude: Exclude | None = EXCLUDE,
    override: Override | None = OVERRIDE,
    remap: Remapping | None = REMAP,
    spaces: bool = SPACES,
    verbose: bool = VERBOSE,
) -> None:
    """Calculate engraving defaults from glyph contours (Python API).

    :param font: Target :class:`.Font` object or path to file.
    :param exclude: :class:`.EngravingDefaults` attributes to exclude.
        Defaults to :obj:`None`.
    :param override: :class:`.EngravingDefaults` attributes to manually
        override mapped to their values. Defaults to :obj:`None`.
    :param remap: :class:`.EngravingDefaults` attributes mapped to
        remappings :class:`dict`, e.g.:

        .. code-block:: python

            r = {
                "arrowShaftThickness": {
                    "ruler": "glyphBoundsXMinAbs",
                    "glyph": "uniEB60"
                }
            }

        Defaults to :obj:`None`.
    :param spaces: Whether values for overrides are given in staff
        spaces instead of font units. Defaults to :obj:`False`.
    :param verbose: Make output verbose. Defaults to :obj:`False`.
    :raises TypeError: If any parameter value is not the expected type.
    :raises ValueError: If any parameter value item is not the expected
        type or value.

    """

    print("Starting...")

    font = _normalizeFont(font)
    exclude = _normalizeExclude(exclude)
    override = _normalizeOverride(override)
    remap = _normalizeRemap(remap)

    font.smufl.spaces = False

    iterator = ENGRAVING_DEFAULTS_MAPPING.items()
    if not verbose:
        iterator = tqdm(ENGRAVING_DEFAULTS_MAPPING.items())

    stdUtils.verbosePrint("\nSetting attributes:", verbose)
    for key, mapping in iterator:
        if exclude and key in exclude:
            continue

        if key == "textFontFamily":
            if override and key in override:
                value = override[key]
                setattr(font.smufl.engravingDefaults, key, value)
                stdUtils.verbosePrint(f"\t'{key}': {value}", verbose)
            continue

        rulerName = mapping.get("ruler", "")
        glyphName = mapping.get("glyph", "")
        remapping = remap.get(key, {}) if remap else {}
        rulerName = remapping.get("ruler", rulerName)
        ruler: RulerType = DISPATCHER[rulerName]
        glyphName = remapping.get("glyph", glyphName)

        rulerValue = _getValue(
            font=font,
            key=key,
            glyphName=glyphName,
            ruler=ruler,
            verbose=verbose,
        )
        if rulerValue is None:
            continue

        if override and key in override:
            value = override[key]
        else:
            value = rulerValue
        if spaces and isinstance(value, (int, float)):
            value = font.smufl.toUnits(value)

        setattr(font.smufl.engravingDefaults, key, value)
        if spaces and value and isinstance(value, (int, float)):
            value = font.smufl.toSpaces(value)
        stdUtils.verbosePrint(f"\t'{key}': {value}", verbose)
    stdUtils.verbosePrint("\nSaving font...", verbose)
    font.save()
    print("\nDone!")


def main() -> None:
    """Command line entry point."""
    args = _parseArgs()
    calculateEngravingDefaults(
        args.font,
        exclude=args.exclude,
        override=args.override,
        remap=args.remap,
        spaces=args.spaces,
        verbose=args.verbose,
    )


# -------
# Helpers
# -------


def _normalizeExclude(exclude: Exclude | None) -> Exclude | None:
    # Noralize `exclude`.
    if exclude is None:
        return None

    error.validateType(exclude, (tuple, list), "exclude")

    for item in exclude:
        error.suggestValue(
            item, list(ENGRAVING_DEFAULTS_MAPPING.keys()), "exclude", items=True
        )
    return exclude


def _normalizeOverride(override: Override | None) -> Override | None:
    # Normalize override dictionary.
    if override is None:
        return None

    error.validateType(override, dict, "override")
    for key, value in override.items():
        attributes = list(ENGRAVING_DEFAULTS_MAPPING.keys())
        error.suggestValue(key, attributes, "override", items=True)
        normalizers.normalizeEngravingDefaultsAttr(key, value)

    return override


def _normalizeRemap(remap: Remapping | None) -> Remapping | None:
    # Normalize remap dictionary.
    if remap is None:
        return None

    error.validateType(remap, dict, "remap")

    for key, value in remap.items():
        error.suggestValue(
            key, list(ENGRAVING_DEFAULTS_MAPPING.keys()), "remap", items=True
        )

        for k, v in value.items():
            error.suggestValue(
                k,
                ("ruler", "glyph"),
                f"'remap'['{value}']",
                items=True,
            )

            error.validateType(
                v, (str, int, type(None)), f"'remap'['{value}']", items=True
            )

    return remap


def _getValue(
    font: Font,
    key: str,
    glyphName: str,
    ruler: RulerType,
    verbose: bool,
) -> int | float | None:
    # Get value from ruler function and print error message.
    try:
        glyph = font[glyphName]
        return ruler(glyph)
    except KeyError:
        stdUtils.verbosePrint(
            "Skipping attribute assigned to non-"
            f"existent glyph: '{key}' ('{glyphName}')",
            verbose,
        )
        return None


def _parseArgs() -> argparse.Namespace:
    # Parse command line arguments and options.
    parser = argparse.ArgumentParser(
        description=stdUtils.getSummary(calculateEngravingDefaults.__doc__),
        parents=[
            cli.commonParser(
                "font", exclude=EXCLUDE, spaces=SPACES, verbose=VERBOSE, addHelp=False
            )
        ],
        formatter_class=cli.createHelpFormatter(
            ("ArgumentDefaultsHelpFormatter", "RawDescriptionHelpFormatter")
        ),
    )

    parser.add_argument(
        "-o",
        "--override",
        default=OVERRIDE,
        type=json.loads,
        help=textwrap.dedent(
            """JSON string of attributes and values to manually override
            in the format: '{"<attributeName>": <value>, ...}'"""
        ),
    )

    parser.add_argument(
        "-r",
        "--remap",
        default=REMAP,
        type=json.loads,
        help=textwrap.dedent(
            """JSON string of ruler and glyph remappings
            in the format: '{"<attributeName>": {"ruler": "<functionName>", 
            "glyph": "<glyphName>"}, ...}'"""
        ),
    )
    return parser.parse_args()


if __name__ == "__main__":
    main()
